// CStorage.h
//**********************************************************************************************
//
//
//				(C) Copyright 1998 by Dipl. Phys. Joerg Plewe, HARDCODE Development
//					All rights reserved. Copying, modification,
//					distribution or publication without the prior written 
//					consent of the author is prohibited.
//
//
//	Filename:		CStorage.h
//	Author:			Herkules
//  Created:		199810xx 
//
//	Module:			Storage
//  Dependecies:	STL, create a DLL
//
//  History:
//  $Date$		$Who$		$What$
//	19981220	Herkules	GetValue() now receives a <type>*& instead of the pointer value
//							If the pointer is NULL, storage tries to set the pointer to address
//							of the internal data
//	19981221	Herkules	New iterator semantic: Objects maybe fetched (List::First,Next; Array::Get)
//							without actually loading the data into memory. Loading is performed by 
//							CStoredObject::Refresh() then. This was done to have better control 
//							on object loading and to prepare a 'reference' data type. The old semantic 
//							immediatly loading data still works without change in syntax 
//	19981223	Herkules	Added datatype LongInt
//	19981225	Herkules	Each class gets a unique number within a library. Maybe useful for a 
//							compact reference to a class when storing object references
//	19981226	Herkules	Introduced interface for object references. Currently only for arrays of 
//							references
//	19981226	Herkules	'List' of classes in CStorage recoded to be a std::vector<>; 
//							CStorage::GetClass( CLASSID _ID ) added
//	19990203	Herkules	Introduced CStoredObject::ReAllocate()
//
/*	Purpose:

This is the main header for the storage DLL containing all exported classes.
These are:

	CStoredClass
	CStoredClassList
	CStoredVar
	CStoredVarList
	CStoredObject
	CStoredObjectList
	CStoredObjectArray
	CStorageEvent
	CStorage

Uses STL

*/
//**********************************************************************************************



#pragma once


#ifndef _H_CStorage
#define _H_CStorage







// STL headers
#include <memory>
#include <list>
#include <vector>
#include <string>

// we use STL by within its namespace
using namespace std;




#ifdef EXPORT_STORAGE
#ifndef STORAGE_DONT_INCLUDE_WINDOWS_H
#include <windows.h>
#endif
#endif

//disable warnings on 255 char debug symbols
#pragma warning (disable : 4786)
//disable warnings on extern before template instantiation
#pragma warning (disable : 4231)
//disable warning on missising dll-interface for iterator definition
#pragma warning (disable : 4251)



#ifdef EXPORT_STORAGE
#	define EXPIMP_TEMPLATE extern
#else
#	define EXPIMP_TEMPLATE
#endif



#include "IStorage.h"

#define STORAGE_CLASS


//-------------------------------------------------------------------------------------------------
//
// additional typedefs
//

typedef unsigned char  CLASSNAMELEN;
typedef unsigned char  VARNAMELEN;
typedef unsigned int   OBJECTSIZE;







// this is merely done to avoid nasty compiler warnings.
// see MSDN article 
// HOWTO: Exporting STL Components Inside & Outside of a Class
// Article ID: Q168958  
class CStoredClass;
class CStoredVar;
class CStoredObject;
class CStorageEvent;

//EXPIMP_TEMPLATE template class STORAGE_CLASS std::list<CStoredClass*>;
//EXPIMP_TEMPLATE template class STORAGE_CLASS std::list<CStoredVar*>;






//
// defines
//
#define MAXCLASSNAMELEN (256)
#define MAXVARNAMELEN (256)



// ----------------------------------------------------------------------------
// this template define the access scheme of a list 
//

template <class T> class TICollectionList
{
public:
	virtual void Release()	{ delete this; }
	virtual T*	First() = 0;
	virtual T*	Next() = 0;
	virtual int	Count() = 0;
protected:
	virtual ~TICollectionList() {};
};

//EXPIMP_TEMPLATE template class STORAGE_CLASS TICollectionList<CStoredVar>;
//EXPIMP_TEMPLATE template class STORAGE_CLASS TICollectionList<CStoredClass>;
//EXPIMP_TEMPLATE template class STORAGE_CLASS TICollectionList<CStoredObject>;






// ----------------------------------------------------------------------------
// this template define the access scheme of an random access array 
//

template <class T> class TICollectionArray
{
public:
	virtual void Release()	{ delete this; }
	virtual int	Count() = 0;
	virtual T* Get( int _Index ) = 0;
	virtual T* operator[] ( int _Index ) { return Get( _Index ); }
protected:
	virtual ~TICollectionArray() {};
};

//EXPIMP_TEMPLATE template class STORAGE_CLASS TICollectionArray<CStoredVar>;
//EXPIMP_TEMPLATE template class STORAGE_CLASS TICollectionArray<CStoredClass>;
//EXPIMP_TEMPLATE template class STORAGE_CLASS TICollectionArray<CStoredObject>;





//-------------------------------------------------------------------------------------------------
//
// Global functions
//

extern POP CreatePOP();








//-------------------------------------------------------------------------------------------------
// Herkules
// CStoredVar
// Contains the type information for a variable to be stored
//

class STORAGE_CLASS CStoredVar : public IStoredVar
{
	friend class CStoredClass;
public:

protected:

	// only CStoredClass may create
	CStoredVar( const string& _Varname, const STOREDVAR_TYPE _Type, const STOREDVAR_ID _VarID ) :
	IStoredVar( _Varname, _Type, _VarID )
	{}

	virtual ~CStoredVar() {}

private:

};









// ----------------------------------------------------------------------------
// simple list object used to enumerate StoredVars

class STORAGE_CLASS CStoredVarList : public IStoredVarList // TICollectionList<CStoredVar>
{
public:

	// fins the first variable
	virtual IStoredVar*	First()	
	{ 
		m_It = m_List.begin();
		if( m_It != m_List.end() )	return (*m_It);
		return NULL;
	}


	// after First(), find the next variable in list 
	virtual IStoredVar*	Next()
	{
		if( ++m_It != m_List.end() )	return (*m_It);
		return NULL;
	}


	// number of variables available
	virtual int	Count() 
	{
		return m_List.size();
	}


private:

	// only CStoredClass may create this list
	friend class CStoredClass;

	// construct
	CStoredVarList( list<CStoredVar*>& _List )	: m_List( _List ) {}
	
	// reference to list of all variables in the class
	list<CStoredVar*>& m_List;

	// iterator
	list<CStoredVar*>::iterator m_It;
};







// ----------------------------------------------------------------------------
// Herkules
// class for object access in an array manner
//

class STORAGE_CLASS CStoredObjectArray : public IStoredObjectArray // TICollectionArray<CStoredObject>
{

public:
	// I am just here! 
	// I inherit pure virtual function, so please derive!

	virtual IStoredObject* Get( int _Index ) 
	{
		return Get( _Index, true );
	}

	// add a method with an additional flag, saying wether the object
	// has to be filled with data immediately
	virtual IStoredObject* Get( int _Index, bool _WithData ) = 0;

};






// ----------------------------------------------------------------------------
// Herkules
// class for object access in an array manner
//

class STORAGE_CLASS CStoredObjectList : public IStoredObjectList // TICollectionList<CStoredObject>
{
public:

	
	// define to new pure virtual function with an additional parameter
	virtual IStoredObject* First( bool _WithData ) = 0;
	virtual IStoredObject* Next( bool _WithData ) = 0;

};





// ----------------------------------------------------------------------------
// defines the type of an object to be stored
// can be accessed by name later
//

class STORAGE_CLASS CStoredClass : public IStoredClass
{
	friend class CStorage;
	friend class CStoredClassList;


public:
	
	// Add a new variable to the class
	virtual IStoredVar* AddVar( const string& _Varname, STOREDVAR_TYPE _Type );

	// retrieve a variable by name
	virtual IStoredVar* GetVar( const string& _Varname );

	// explore the list of variables
	virtual IStoredVarList* GetVarList()	{ return new CStoredVarList( m_lVars ); }

	
	// fire an event that class-objects have changed content
	virtual void FireChangedEvent( IStoredObject* _pObject = NULL,  IStoredVar* _pVar = NULL );


protected:

	// only CStorage and derived can create this object
	CStoredClass(	const string _Classname, 
					const STOREDCLASS_ID _ClassID, 
					IStorageEngine* _pStorage, 
					IStoredClass* const _pBaseClass = NULL );

	virtual ~CStoredClass();

	// create a new variable with given name, type and ID
	// this is used when restoring the variable set from an archive
	virtual CStoredVar* CreateVar( const string& _Varname, STOREDVAR_TYPE _Type, STOREDVAR_ID _ID );


	// add an existing variable to the repository
	void AddVarToRepository( CStoredVar* _pVar );

	// Create a new variable ID
	int NewVarID();

private:
	
	//
	// variables
	//

	// Always remember last variable ID assigned
	// If there is a base class, new variable ID will be obtained from the base class so
	// that m_LastVarID is never used in this case
	STOREDVAR_ID m_LastVarID;

	// keep variables in a list
	list<CStoredVar*> m_lVars;

};



// ----------------------------------------------------------------------------
class STORAGE_CLASS CStoredObject : public IStoredObject
{
	friend class CStorage;
public:



	//
	// fire an event that object has changed content
	//
	virtual void FireChangedEvent( IStoredVar* _pVar = NULL );




protected:
	// default constructor
	CStoredObject( int _ObjectIndex, IStoredClass* _pStoredClass ) : 
	IStoredObject( _ObjectIndex, _pStoredClass )
	{}
private:
}; // StoredObject








//-----------------------------------------------------------------------------CStoredClassList----
// Herkules
// simple list object used to enumerate StoredClasses
//

class STORAGE_CLASS CStoredClassList : public IStoredClassList // TICollectionList<CStoredClass>
{

public:

	// retrieve first element
	virtual IStoredClass* First()	
	{ 
		m_It = m_List.begin();
		if( m_It != m_List.end() )		return *m_It;
		return NULL;
	}

	// continously get next
	virtual IStoredClass* Next()
	{
		if( ++m_It != m_List.end() )	return *m_It;
		return NULL;
	}

	// how many are there?
	virtual int	Count() 
	{
		return m_List.size();
	}


private:

	// only CStorage may create
	friend class CStorage;

	// Constructor
	CStoredClassList( const vector<IStoredClass*>& _List )	: m_List( _List ) {}
	
	// reference to the list of classes
	const vector<IStoredClass*>& m_List;

	// iterator for First() and Next()
	vector<IStoredClass*>::const_iterator m_It;

}; // CStoredClassList







// ----------------------------------------------------------------------------
// event class
// has to be derived
//

class STORAGE_CLASS CStorageEvent : public IStorageEvent
{

public:

	// Fire the event
	virtual void Fire( IStoredClass* _pClass = NULL , IStoredObject* _pObject = NULL, IStoredVar* _pVar  = NULL ) = 0;

	// Fire event without invoking Fire
	void operator()( IStoredClass* _pClass = NULL, IStoredObject* _pObject = NULL, IStoredVar* _pVar = NULL )
	{
		Fire( _pClass, _pObject, _pVar );
	}

};









//-------------------------------------------------------------------------------------------------
// the class to start with
//

class STORAGE_CLASS CStorage : public IStorageEngine
{

public:

	// Constuctor
	CStorage( IStorageEvent* _pEvent = NULL );
	//static CStorage* New( int _StorageType = Storagetype::Memory, CStorageEvent* _pEvent = NULL );


	// Destructor
	virtual ~CStorage();


	//
	// merge with the repository from another database
	//
	bool MergeRepository( const CStorage& _Source );

	//
	// copy the data content from another database 
	// 
	bool CopyContent( const CStorage& _Source );

	//
	// read data from a special text file
	//
	virtual bool ReadTextFile( const char* _Filename );


	//
	// write data to a special text file
	//
	virtual bool WriteTextFile( const char* _Filename );




	//
	// close the database
	//
	virtual bool Close();							// save index and close all files


	//
	// fire an event that class-objects have changed content
	//
	virtual void FireChangedEvent( IStoredClass* _pClass = NULL, IStoredObject* _pObject = NULL, IStoredVar* _pVar = NULL );




	// maintain structure
	

	//
	// add a new class to the repository
	//
	virtual IStoredClass* AddClass( const string& _Classname, IStoredClass* const _pBaseClass = NULL );


	//
	// find a class in repository by name
	//
	virtual IStoredClass* GetClass( const string& _Classname ) const;

	//
	// find a class in repository by ID
	//
	virtual IStoredClass* GetClass( const STOREDCLASS_ID _ClassID ) const;



	//
	// get an access object to find out about all classes
	// 
	virtual IStoredClassList* GetClassList() const { return new CStoredClassList( m_lClasses ); }


protected:
	//
	// used by derived class in order to create correct CStoredClass-derived object
	//
	virtual CStoredClass* CreateClass(	const string& _ClassName, 
										const STOREDCLASS_ID _ClassID, 
										IStoredClass* const _pBaseClass = NULL ) = 0;
	
	//
	// add an existing class
	//
	void AddClassToRepository( CStoredClass* _pNewClass );

	//
	// given the name, find the class
	//
	CStoredClass* FindClassInRepository( const string& _Classname ) const;

private:

	// CStoredClasses are stored in a list
	vector<IStoredClass*> m_lClasses;

	// just remember of what kind I am
	int m_StorageImplementation;

	// pointer to the event object
	IStorageEvent* m_pStorageEvent;

}; // CStorage




#undef EXPIMP_TEMPLATE
#endif // _H_CStorage


