// IStorage.h
//**********************************************************************************************
//
//
//				(C) Copyright 1998 by Dipl. Phys. Joerg Plewe, HARDCODE Development
//					All rights reserved. Copying, modification,
//					distribution or publication without the prior written 
//					consent of the author is prohibited.
//
//
//	Filename:		IStorage.h
//	Author:			Herkules
//  Created:		19990205
//  Version:		$Version$
//
//	Module:			Storage
//  Dependecies:	$Dependencies$
//
//  History:
//  $Date$		$Who$		$What$
//	19990317	Herkules	Added IStoredClass::IsDerived()
//
/*	Purpose:

Interfaces to the storage module

This header file describes the interface of the module 'Storage'. 




*/
//**********************************************************************************************

#pragma once


#ifndef _H_IStorage
#define _H_IStorage


//-------------------------------------------------------------------------------------------------
// symbols that may be defined for the preprocessor
//
// USE_STORAGE_WITHOUT_DLL 
//
// If is is defined, including the header does not cause any open references to code,
// so no lib has to be linked. This is useful for using the Storage in submodules of an 
// application, receiving their Storage pointer from elsewhere
//




//disable warning on missising dll-interface for iterator definition
#pragma warning (disable : 4251)


#include <string>
// we use STL by within its namespace
using namespace std;




#ifdef EXPORT_STORAGE
#	define STORAGE_CLASS __declspec(dllexport)
#else
#	define STORAGE_CLASS __declspec(dllimport)
#endif



//-------------------------------------------------------------------------------------------------
// Herkules
// Interfaces to be defined:
//

class IStorageEngine;
class IStoredVar;
class IStoredClass;
class IStoredObject;
class IStorageEvent;


// iterators
class IStoredVarList;
class IStoredObjectList;
class IStoredObjectArray;
class IStoredClassList;






//-------------------------------------------------------------------------------------------------
//
// typedefs
//

typedef unsigned short STOREDVAR_ID;
typedef unsigned short STOREDVAR_TYPE;
typedef unsigned short STOREDCLASS_ID;


//
// typedefs for the internal types
//
typedef unsigned char	Byte;
typedef short			Short;
typedef int				Int;
typedef __int64			LongInt;
typedef float			Float;
typedef double			Double;

// Persistent Object Pointer
typedef __int64			POP;




//-------------------------------------------------------------------------------------------------
//
// array variables have the highest bit set in the vartype
//

// All array types have a certain bit set
#define VARTYPEARRAYMASK			0x8000

// Check wether a variable type is an array
#define ISARRAYVAR( VarType )		( VarType & VARTYPEARRAYMASK )

// From a variable type, get the scalar base type
#define STORED_BASETYPE( VarType )	( VarType & ~VARTYPEARRAYMASK )



//-------------------------------------------------------------------------------------------------
//
// Global functions
//
#ifndef USE_STORAGE_WITHOUT_DLL


//-----------------------------------------------------------------------------CreatePOP-----------
// Herkules
// Create a 64bit random number that can be used to universal addressing
// This is near to a GUID, mit more compact and perfectly supported by the 
// compiler (__int64)
//

extern STORAGE_CLASS POP CreatePOP();

#endif





//-----------------------------------------------------------------------------IStorageEngine------
// Herkules
// The class to start with
//
// The Engine itself. An implementation of this interface represents a Storage driver. 
// Instances are create with the static New() or can be new-ed directly, when the driver
// has a public constructor.
//

class STORAGE_CLASS IStorageEngine
{

public:

	enum 
	{
		Memory = 1,
		Library,
		Internet,		// currently not implemented
		NumberOfStorageImplementations
	};


#	ifndef USE_STORAGE_WITHOUT_DLL

	// Static function creating an instance for one of the drivers currently implemented
	static IStorageEngine*		New( int _StorageType = IStorageEngine::Memory, IStorageEvent* _pEvent = NULL );

#	endif
	
	// Connect to a certain datasource, e.g. a file or internet or ....
	virtual bool				Connect( const char* _Connection ) 
								{ return true; }

	// Close the database and free all resources
	virtual bool				Close() = 0;

	// Read data from a special text file
	virtual bool				ReadTextFile( const char* _Filename )	
								{ return false; }

	// Write data to a special text file
	virtual bool				WriteTextFile( const char* _Filename ) 
								{ return false; }

	// Fire an event that class-objects have changed content
	virtual void				FireChangedEvent(	IStoredClass* _pClass = NULL, 
													IStoredObject* _pObject = NULL, 
													IStoredVar* _pVar = NULL ) = 0;

	//
	// Maintain structure
	//

	// Add a new class to the repository
	virtual IStoredClass*		AddClass( const string& _Classname, IStoredClass* const _pBaseClass = NULL ) = 0;

	// Find a class in repository by name
	virtual IStoredClass*		GetClass( const string& _Classname ) const = 0;

	// Find a class in repository by ID
	virtual IStoredClass*		GetClass( const STOREDCLASS_ID _ClassID ) const = 0;

	// Get an access object to find out about all classes
	virtual IStoredClassList*	GetClassList() const = 0;

	// Remove the driver. Engine is no longer operational
	virtual bool				Release()		
								{ delete this; return true; }

protected:

	// Destructor
	virtual						~IStorageEngine() 
								{}

}; // IStorageEngine



//-----------------------------------------------------------------------------IStoredVar----------
// Herkules
// IStoredVar
// Contains the type information for a variable to be stored
//

class STORAGE_CLASS IStoredVar
{

public:

	//
	// Define all datatypes a variable may have
	//
	enum 
	{
		Byte = 1,
		Short,
		Int,
		LongInt,
		Float,
		Double,
		String,
		Reference,

		NumberOfScalarTypes,

		// array types have the most significant bit set
		// THIS IS AN IMPORTANT CONVENTION
		ByteArray = VARTYPEARRAYMASK + 1,
		ShortArray,
		IntArray,
		LongIntArray,
		FloatArray,
		DoubleArray,
		StringArray,		// not supported!!!!!
		ReferenceArray,
		
		NumberOfTypes
	};

	
	// Retrieve or set the name of a variable
	const string&				Name() const 
								{ return m_Varname; }

	// Retrieve the type
	const STOREDVAR_TYPE&		GetType() const 
								{ return m_Vartype; }

	// Retrieve the ID
	const STOREDVAR_ID&			GetID()	const 
								{ return m_VarID; }


protected:

	
	// Only friends may create
	// Get your IStoredVar from a IStoredClass!
								IStoredVar( const string& _Varname, 
											const STOREDVAR_TYPE _Type, 
											const STOREDVAR_ID _VarID ) :
								m_Vartype( _Type ),
								m_VarID( _VarID ),
								m_Varname( _Varname )
								{}


	// Destructor
	virtual						~IStoredVar()		
								{}

	// Each variable has unique ID within its class
	const STOREDVAR_ID			m_VarID;

	// And a defined datatype that cannot be changed later
	const STOREDVAR_TYPE		m_Vartype;

	// The name of the variable
	const string				m_Varname;

};	// IStoredVar












//-----------------------------------------------------------------------------IStoredClass--------
// Herkules
// Defines the type of an object to be stored.
// Can be accessed by name later
//

class STORAGE_CLASS IStoredClass 
{

public:
	
	// Allow access to the storage that owns the class.
	// It is returned as a pointer instead of as reference to allow casts to 
	// derived classes.
	virtual IStorageEngine*		GetStorage()		
								{ return m_pStorage; }

	// Add a new variable to the class
	virtual IStoredVar*			AddVar( const string& _Varname, STOREDVAR_TYPE _Type ) = 0;

	// Retrieve a variable by name
	virtual IStoredVar*			GetVar( const string& _Varname ) = 0;

	// Explore the list of variables
	virtual IStoredVarList*		GetVarList() = 0;

	// Add a new object
	virtual IStoredObject*		AddObject() = 0;

	// Get a retrieval object with array access
	virtual IStoredObjectArray*	GetObjectArray() = 0;

	// Get a retrieval object with list access
	virtual IStoredObjectList*	GetObjectList() = 0;
	
	// Get the name of the class
	const string&				Name() const 
								{ return m_Classname; }

	// Retrieve the ID
	const STOREDCLASS_ID		GetID() const 
								{ return m_ClassID; }

	// Fire an event that class-objects have changed content
	virtual void				FireChangedEvent( IStoredObject* _pObject = NULL, IStoredVar* _pVar = NULL ) = 0;

	IStoredClass* const			GetBaseClass()
								{ return m_pBaseClass; }

	bool						IsDerived( IStoredClass* _pBaseClass )
	{
		// Get the ID of the target baseclass. It can be used for comparison of classes
		int BaseClassID = _pBaseClass->GetID();

		// Loop until either the pTestClass has the desired type or there are no more base classes
		while(		_pBaseClass->GetID() != BaseClassID  
				&&  (_pBaseClass = _pBaseClass->GetBaseClass()) )
		{}

		// Has a suitable class been found?
		return _pBaseClass != NULL; 
	}

protected:

	// Constructor
	// Only friends may create an IStoredClass object. Use GetClass() or AddClass() of the 
	// IStorageEngine to ask for classes.
								IStoredClass::IStoredClass( const string _Classname, 
															const STOREDCLASS_ID _ClassID, 
															IStorageEngine* _pStorage,
															IStoredClass* const _pBaseClass = NULL ) :
								m_pStorage( _pStorage ),
								m_Classname( _Classname ),
								m_ClassID( _ClassID ),
								m_pBaseClass( _pBaseClass )
								{}


	// Destructor
	virtual						~IStoredClass() 
								{}


	//
	// Variables
	//

	// The name of the class
	string						m_Classname;

	// The unique number of a class within a storage
	STOREDCLASS_ID				m_ClassID;

	// Keep a pointer to the actual storage the class belongs to.
	// This has to be a pointer in spite of a reference, bc. it has to be
	// casted to derived classes for the drivers.
	IStorageEngine*				m_pStorage;

	// Pointer to the baseclass we derive from
	IStoredClass* const m_pBaseClass;
	

};	// IStoredClass














//-----------------------------------------------------------------------------IStoredObject-------
// Herkules
// The data container itself
//
// An IStoredObject is the main workhorse of the Storage system. It contains all content data 
// an manages the access. 
// It provides semanticly rich, typesafe and, last but not least, fast access to its content for 
// object setup (AddValue), data retrieval (GetValue) and data maintenance (SetValue) 
//

class STORAGE_CLASS IStoredObject
{

public:

	// Allow access to the IStoredClass that owns the object.
	// It is returned as a pointer instead of as reference to allow casts to 
	// derived classes
	IStoredClass*				GetClass()		
								{ return m_pStoredClass; }


	// check wether a certain value is present
	virtual bool				HasValue( IStoredVar* _pVar ) = 0;




	//
	// Add data
	//
	virtual bool				AddValue( IStoredVar* _pVar, unsigned char _Value )					= 0;
	virtual bool				AddValue( IStoredVar* _pVar, short _Value )							= 0;
	virtual bool				AddValue( IStoredVar* _pVar, int _Value )							= 0;
	virtual bool				AddValue( IStoredVar* _pVar, LongInt _Value )						= 0;
	virtual bool				AddValue( IStoredVar* _pVar, float _Value )							= 0;
	virtual bool				AddValue( IStoredVar* _pVar, double _Value )						= 0;
	virtual bool				AddValue( IStoredVar* _pVar, const string& _Value, int _MaxSize = 0 ) = 0;

	//
	// Add arrays
	//
	virtual bool				AddValue( IStoredVar* _pVar, const unsigned char* _pValue, int _Size = 1 ) = 0;
	virtual bool				AddValue( IStoredVar* _pVar, const short* _pValue, int _Size = 1 )		= 0;
	virtual bool				AddValue( IStoredVar* _pVar, const int* _pValue, int _Size = 1 )		= 0;
	virtual bool				AddValue( IStoredVar* _pVar, const LongInt* _pValue, int _Size = 1 )	= 0;
	virtual bool				AddValue( IStoredVar* _pVar, const float* _pValue, int _Size = 1 )		= 0;
	virtual bool				AddValue( IStoredVar* _pVar, const double* _pValue, int _Size = 1 )		= 0;
//	virtual bool				AddValue( IStoredVar* _pVar, const string* _pValue, int _Size = 1, int _MaxSize = 0 ) = 0;
	
	
	//
	// Retrieve data
	//
	virtual bool				GetValue( IStoredVar* _pVar, unsigned char& _Value )		= 0;
	virtual bool				GetValue( IStoredVar* _pVar, short& _Value )				= 0;
	virtual bool				GetValue( IStoredVar* _pVar, int& _Value )					= 0;
	virtual bool				GetValue( IStoredVar* _pVar, LongInt& _Value )				= 0;
	virtual bool				GetValue( IStoredVar* _pVar, float& _Value )				= 0;
	virtual bool				GetValue( IStoredVar* _pVar, double& _Value )				= 0;
	virtual bool				GetValue( IStoredVar* _pVar, string& _Value )				= 0;


	//
	// Get arrays 
	//
	virtual bool				GetValue( IStoredVar* _pVar, unsigned char*& _pValue, int& _Size )	= 0;
	virtual bool				GetValue( IStoredVar* _pVar, short*& _pValue, int& _Size )			= 0;
	virtual bool				GetValue( IStoredVar* _pVar, int*& _pValue, int& _Size )			= 0;
	virtual bool				GetValue( IStoredVar* _pVar, LongInt*& _pValue, int& _Size )		= 0;
	virtual bool				GetValue( IStoredVar* _pVar, float*& _pValue, int& _Size )			= 0;
	virtual bool				GetValue( IStoredVar* _pVar, double*& _pValue, int& _Size )			= 0;

	//
	// Modify data
	//
	virtual bool				SetValue( IStoredVar* _pVar, unsigned char _Value )		= 0;
	virtual bool				SetValue( IStoredVar* _pVar, short _Value )				= 0;
	virtual bool				SetValue( IStoredVar* _pVar, int _Value )				= 0;
	virtual bool				SetValue( IStoredVar* _pVar, LongInt _Value )			= 0;
	virtual bool				SetValue( IStoredVar* _pVar, float _Value )				= 0;
	virtual bool				SetValue( IStoredVar* _pVar, double _Value )			= 0;
	virtual bool				SetValue( IStoredVar* _pVar, const string& _Value )		= 0;

	//
	// Modify arrays 
	//
	virtual bool				SetValue( IStoredVar* _pVar, unsigned char* _pValue, int& _Size ) = 0;
	virtual bool				SetValue( IStoredVar* _pVar, short* _pValue, int& _Size )		= 0;
	virtual bool				SetValue( IStoredVar* _pVar, int* _pValue, int& _Size )			= 0;
	virtual bool				SetValue( IStoredVar* _pVar, LongInt* _pValue, int& _Size )		= 0;
	virtual bool				SetValue( IStoredVar* _pVar, float* _pValue, int& _Size )		= 0;
	virtual bool				SetValue( IStoredVar* _pVar, double* _pValue, int& _Size )		= 0;

	//
	// References to other objects
	//

	// References are always reference arrays
	virtual bool				GetValue( IStoredVar* _pVar, IStoredObject** _pObj, int& _Size )	= 0; 
	virtual bool				AddValue( IStoredVar* _pVar, IStoredObject** _pObj, int _Size = 1 )	= 0; 
	virtual bool				SetValue( IStoredVar* _pVar, IStoredObject** _pObj, int& _Size )	= 0; 

	// Fire an event that object has changed content
	virtual void				FireChangedEvent( IStoredVar* _pVar = NULL ) = 0;



	//
	// Maintain the object
	//

	// Flush the object to the storage medium.
	// After creating an object with IStoredClass::AddObject() and adding some values to the new
	// object, calling Flush() makes to object operational.
	// Before doing the initial Flush(), an object cannot be used with GetValue() or SetValue(), 
	// nor the index returned by GetIndex() has to have a meaning nor can the object be used 
	// as the content for reference variable. 
	// In fact, all you can do is AddValue() and Flush().
	//
	// After the initial call, Flush() has to be reapplied to make changes made with SetValue() 
	// valid and visible to other clients of the same object. Flush() may create activity on
	// the storage medium.
	// The _KeepData flag advises the IStoredObject wether or not to free the internal content.
	// If the content is released, GetValue() will always return 'false' until Refresh() is called 
	// to reload an object. If _KeepData is 'true', the object stays fully operational
	virtual bool				Flush( bool _KeepData = false ) = 0;

	// Reload a Flushed object and fill in its content 
	virtual bool				Refresh()						
								{ return true; }

	// Release the object and its content. This is NOT implemented with 'delete this;'
	// as a default implementation, bc. for example the objects in the 'memory' driver
	// actually never may be deleted during the liftime of the engine 
	virtual bool				Release()
								{ return true; }

	// Get the index of the object usable for array access
	const int					GetIndex()	
								{ return m_ObjectIndex; }

	// Restart the object, free and reallocate the content.
	// After ReAllocate(), the object is in nearly the same state as immediately after 
	// IStoredClass::AddObject(), except that the GetIndex() value is still valid and 
	// garuanteed to be the same as before.
	// All references to the object will stay valid.
	virtual bool				ReAllocate()		
								{ return false; }

protected:


								IStoredObject::IStoredObject( int _ObjectIndex, IStoredClass* _pStoredClass ) : 
								m_ObjectIndex( _ObjectIndex ),
								m_pStoredClass( _pStoredClass )
								{}

	// Destructor
	virtual						~IStoredObject() 
								{}

	//
	// Variables
	//

	// The unique index of the object within its class
	int							m_ObjectIndex;

	// An object needs information from the class
	IStoredClass*				m_pStoredClass;

};	// IStoredObject






//-----------------------------------------------------------------------------IStorageEvent-------
// Herkules
// Event class
//
// The class has to be implemented by the user of the Storage. A pointer to an object 
// implementing this interface can be passed when creating an instance of the
// IStorageEngine.
// The Fire() method will be called when one of the FireChanged() methods is executed.
//

class STORAGE_CLASS IStorageEvent
{

public:

	// Virtual destructor
	virtual						~IStorageEvent() 
								{}

	// Fire the event, telling which class, object and variable has be changed.
	// The parameter may be NULL in order to indicate that a larger set has changed.
	// E.g. _pVar=NULL means, that the object as a whole has changed.
	// If also _pObject=NULL, the whole class or a couple of objects changed.
	// If even _pClass=NULL, the entire library may have been changed and applications
	// better should Refresh() all objects in use.
	virtual void				Fire(	IStoredClass* _pClass = NULL , 
										IStoredObject* _pObject = NULL, 
										IStoredVar* _pVar  = NULL ) = 0;

	// Offer the same Release() method some Storage classes have
	virtual bool				Release()	
								{ delete this; return true; }

};	// IStorageEvent










//*************************************************************************************************
//
//
//--Iterators------------------------------------------------------------------Iterators-----------
//
//
//*************************************************************************************************






//-----------------------------------------------------------------------------IStoredVarList------
// Herkules
// Simple list interface used to enumerate StoredVars
//
// It is used as an iterator on the available variables. A pointer to that interface can only
// be obtained from a IStoredClass. 
// After using a IStoredVarList, Release() has to be called on the object in order to free its 
// memory.
//

class STORAGE_CLASS IStoredVarList
{

public:

	// Finds the first variable
	virtual IStoredVar*			First()	= 0;

	// After First(), find the next variable in list 
	virtual IStoredVar*			Next()	= 0;

	// Number of variables available
	virtual int					Count() = 0;

	// Free the object implementing the interface
	virtual void				Release()		{ delete this; }

protected:

	// Destructor
	virtual						~IStoredVarList()	
								{}

};	// IStoredVarList




//-----------------------------------------------------------------------------IStoredObjectList---
// Herkules
// Simple list interface used to enumerate StoredObjects
//
// It is used as an iterator on the available object of a given class. 
// A pointer to that interface can only be obtained from a IStoredClass. 
// After using a IStoredObjectList, Release() has to be called on the object in order to free its 
// memory.
//

class STORAGE_CLASS IStoredObjectList
{
public:

	// Finds the first object, resets the iterator
	virtual IStoredObject*		First( bool _WithData = true ) = 0;

	// After First(), find the next variable in list 
	virtual IStoredObject*		Next( bool _WithData = true ) = 0;

	// Number of objects available
	virtual int					Count() = 0;

	// Delete the list access object
	virtual bool				Release()		
								{ delete this; return true; }

protected:

	// Destructor
	virtual ~IStoredObjectList()	{}

};	// IStoredObjectList






//-----------------------------------------------------------------------------IStoredObjectArray--
// Herkules
// Class for object access in an array manner
//

class STORAGE_CLASS IStoredObjectArray
{
public:

	// Access a certain object
	virtual IStoredObject*		Get( int _Index, bool _WithData = true ) = 0;

	// Number of objects available
	virtual int					Count()	= 0;

	// Delete the array access object
	virtual bool				Release()			
								{ delete this; return true; }

protected:

	// destructor
	virtual						~IStoredObjectArray()	
								{}

};	// IStoredObjectArray







//-----------------------------------------------------------------------------IStoredClassList----
// Herkules
// Simple list interface used to enumerate IStoredClasses
//

class STORAGE_CLASS IStoredClassList
{

public:

	// Finds the first variable
	virtual IStoredClass*		First()	= 0;

	// After First(), find the next IStoredClass in list 
	virtual IStoredClass*		Next() = 0;

	// Number of variables available
	virtual int					Count() = 0;

	// Release the object implementing the interface
	virtual bool				Release()	
								{ delete this; return true; }

protected:

	// Destructor
	virtual						~IStoredClassList()	
								{}

};	// IStoredClassList







#undef STORAGE_CLASS

#endif // _H_IStorage
