// CStorageLibrary.cpp
//**********************************************************************************************
//
//
//				(C) Copyright 1998 by Dipl. Phys. Joerg Plewe, HARDCODE Development
//					All rights reserved. Copying, modification,
//					distribution or publication without the prior written 
//					consent of the author is prohibited.
//
//
//	Filename:		CStorageLibrary.cpp
//	Author:			Herkules
//  Created:		9810xx 
//
//	Module:			StorageLibrary
//  Dependecies:	CStorage and CStorageMemory classes
//
//  History:
//  $Date$		$Who$		$What$
//	19981221	Herkules	Refresh() can allocate the data itself. This was needed for the 
//							new semantic of the iterators, allowing an object fetch without 
//							actually loading the data
//	19990203	Herkules	Implemented CStoredObject::ReAllocate()
//
/*	Purpose:

Implementation of the library storage classes.

*/
//**********************************************************************************************


#define EXPORT_STORAGE
#include "CStorageLibrary.h"

#include <assert.h>



// --------------------------------------------------------------------------------
// helper macro doing a more secure close of a file
//

#define CLOSEHANDLE( Handle )		\
{									\
	CloseHandle( Handle );			\
	Handle = INVALID_HANDLE_VALUE;	\
}




//-------------------------------------------------------------------------------------------------
//
//
//
// CStoredObjectLibrary                                                        CStoredObjectLibrary 
//
//
//
//-------------------------------------------------------------------------------------------------



//-----------------------------------------------------------------------------CStoredObjectLibrary
// Herkules
// Constructor to be used by the iterators or reference resolution
// If the _OffsetData and _CompiledSize are not given, they have to be
// read from the header at _OffsetHeader
//

CStoredObjectLibrary::CStoredObjectLibrary( CStoredClassLibrary* _pStoredClass, 
											int _Index, 
											FILEOFFSET _OffsetHeader, 
											FILEOFFSET _OffsetData, 
											OBJECTSIZE _CompiledSize ) :
super( _Index, _pStoredClass )
{

	
	//
	// set these values so Refresh can fully create the memory layout!
	//
	m_MyFileOffsetHeader = _OffsetHeader;

	m_MyFileOffsetData = _OffsetData;
	m_ObjectSize = _CompiledSize;

}



//-----------------------------------------------------------------------------CStoredObjectLibrary
// Herkules
// Constructor used by AddObject(). It is just a hull unless Flush() is called
//

CStoredObjectLibrary::CStoredObjectLibrary( CStoredClassLibrary* _pStoredClass ) :
super( -1, _pStoredClass )
{
	//
	// the object index given to the super class is -1 bc. we do not 
	// know a valid index until the object is flushed
	// 

	m_MyFileOffsetHeader = 0;
	m_MyFileOffsetData = 0;
	m_ObjectSize = 0;

}




bool CStoredObjectLibrary::SaveDataToFile()
{

	//
	// locals
	//
	CStorageLibrary* pStorage = static_cast<CStorageLibrary*>(GetClass()->GetStorage());
	const HANDLE File = pStorage->GetDatafile();
	DWORD BytesWritten, BytesRead;
	BOOL Ret;
	DWORD Pos;
	bool DataAppended = false;


	// object must already have a header in the file 
	assert( m_MyFileOffsetHeader );

	//
	// does it already have a data area in the file?
	//
	if( m_MyFileOffsetData == 0 )
	{
		// 
		// no data area so far!
		//

		_FILEOBJECTHEADER Header;


		// get position for new object in the file 
		m_MyFileOffsetData = pStorage->EndOfObjects();

		
		// 
		// set objects data in its header
		// 

		// put the file-read pointer to the position of the  object
		Pos = SetFilePointer( File, m_MyFileOffsetHeader, NULL, FILE_BEGIN );
		if( Pos != m_MyFileOffsetHeader )
			return FALSE;

		//
		// read in the header again
		// 
		Ret = ReadFile( File, &Header, sizeof( _FILEOBJECTHEADER ), &BytesRead, NULL );
		if( ! Ret || BytesRead != sizeof( _FILEOBJECTHEADER ) )
			return false;
		
		// set the new header values
		Header.m_Size = FlushedObjectSize();
		Header.m_DataBlock = m_MyFileOffsetData;



		// put the file-write pointer to the position of the object
		Pos = SetFilePointer( File, m_MyFileOffsetHeader, NULL, FILE_BEGIN );
		if( Pos != m_MyFileOffsetHeader )
			return false;

		Ret = WriteFile( File, (LPCVOID)&Header, sizeof( _FILEOBJECTHEADER ), &BytesWritten, NULL );
		if( ! Ret || BytesWritten != sizeof( _FILEOBJECTHEADER ) ) 
			return 0;

		// set flag the we are going to append data
		DataAppended = true;
	}



	// put the file-write pointer to the position of the  object
	// thereby skip the objects header
	Pos = SetFilePointer( File, m_MyFileOffsetData, NULL, FILE_BEGIN );
	if( Pos != m_MyFileOffsetData )
		return FALSE;


	// write the file with the data
	Ret = WriteFile( File, (LPCVOID)FlushedObjectData(), FlushedObjectSize(), &BytesWritten, NULL );
	if( ! Ret || BytesWritten != FlushedObjectSize() ) 
		return false;



	if( DataAppended )
	{

		// repository will have to be rewritten to file
		pStorage->m_bRepositoryDirty = true;


		// update the storage object: where to put next object
		pStorage->EndOfObjects() = m_MyFileOffsetData + FlushedObjectSize();

	}




	// everything went fine
	return true;


}




//-----------------------------------------------------------------------------Flush---------------
// Herkules
// depending on the history - append the object the a file or update
// an existing record
//

bool 
CStoredObjectLibrary::Flush( bool _KeepData )
{
	bool Success = false;

	
	// let the super class flush its object
	// this may fail e.g. when re-Flushing an already flushed object!
	if( super::Flush() )
	{

		// retrieve a pointer to our class
		CStoredClassLibrary* pClass = static_cast<CStoredClassLibrary*>(m_pStoredClass);

		
		//
		// does the object already have a header in the file?
		//
		if( ! m_MyFileOffsetHeader )
		{

			// we ask the class to store the object, bc. it has to maintain class-internal values
			// like where it stored the last object of its kind
			m_MyFileOffsetHeader = pClass->CreateHeaderInFile( m_ObjectIndex );
		}


		//
		// already has a position in file -> object has been stored earlier
		//
		Success = SaveDataToFile();
			
	}


	//
	// If we may get rid of the data now - do it!
	// This is NOT done in the base class. We are in luck!
	//
	if( ! _KeepData )
		DeAllocateCompiledObject();


	// return how we did
	return Success;


} // Flush


/*
//-----------------------------------------------------------------------------Flush---------------
// Herkules
// depending on the history - append the object the a file or update
// an existing record
//

bool 
CStoredObjectLibrary::Flush( bool _KeepData )
{
	bool Success = false;

	
	// let the super class flush its object
	// this may fail e.g. when re-Flushing an already flushed object!
	if( super::Flush() )
	{

		// retrieve a pointer to our class
		CStoredClassLibrary* pClass = static_cast<CStoredClassLibrary*>(m_pStoredClass);

		
		//
		// does the object already have a header in the file?
		//
		if( m_MyFileOffsetHeader )
		{
			//
			// already has a position in file -> object has been stored earlier
			//

			Success = pClass->UpdateObjectData( m_MyFileOffsetHeader, FlushedObjectData(), FlushedObjectSize() );
			
		}
		else
		{
			//
			// needs new storage space in file
			//


			// we ask the class to store the object, bc. it has to maintain class-internal values
			// like where it stored the last object of its kind
			m_MyFileOffsetHeader = pClass->AppendObjectData( FlushedObjectData(), FlushedObjectSize(), m_ObjectIndex );
		
			// if the object is lucky somewhere in the file now, the offset is non-zero
			Success = ( m_MyFileOffset != 0 );

		}

	}


	//
	// If we may get rid of the data now - do it!
	// This is NOT done in the base class. We are in luck!
	//
	if( ! _KeepData )
		DeAllocateCompiledObject();


	// return how we did
	return Success;


} // Flush
*/



// --------------------------------------------------------------------------------
// Herkules
// Re-read object from file
//

bool 
CStoredObjectLibrary::Refresh( )
{
	// locals
	DWORD Pos;
	unsigned long BytesRead;

	// super class refreshes first
	if( ! super::Refresh() )
		return false;



	// not flushed - whatever...
	if( ! m_MyFileOffsetHeader )		
		return false;


	// find the handle of the file from the CStorage object
	HANDLE hDatFile = static_cast<CStorageLibrary*>(GetClass()->GetStorage())->GetDatafile();


	
	// is the objects memory is not yet there, allocate it now!
	if( ! FlushedObjectData() )	//return false;
	{
		unsigned char * pData;

		//
		// if the object has been created without giving the size, read the size now from file
		//
		if( m_ObjectSize == 0 )
		{
			_FILEOBJECTHEADER Header;

			// put read pointer to begin of object 
			Pos = SetFilePointer( hDatFile, m_MyFileOffsetHeader, NULL, FILE_BEGIN );

			// if not at start of object, return
			if( Pos != m_MyFileOffsetHeader)
				return false;

			// read the data
			if( ! ReadFile( hDatFile, &Header, sizeof( _FILEOBJECTHEADER ), &BytesRead, NULL ) )	
				return false;

			// get the size from the header
			m_ObjectSize = Header.m_Size;
			m_MyFileOffsetData = Header.m_DataBlock;
		}


		// the size given here must be m_ObjectSize, bc. FlushedObjectSize() only works when
		// data is already allocated!
		pData = AllocateCompiledObject( m_ObjectSize );

		// assign buffer  to object
		AssignData( pData, m_ObjectSize );
	}


	// put read pointer to begin of object 
	Pos = SetFilePointer( hDatFile, m_MyFileOffsetData, NULL, FILE_BEGIN );

	// if not at begin of object, return
	if( Pos != m_MyFileOffsetData )
		return false;


	// read the data
	if( ! ReadFile( hDatFile, FlushedObjectData(), FlushedObjectSize(), &BytesRead, NULL ) )	
		return false;

	// everything OK
	return true;


} // Refresh





//-----------------------------------------------------------------------------Release-------------
// Herkules
// coming and going....
//

bool 
CStoredObjectLibrary::Release( )
{
	// DeAllocateCompiledObject() is done when deleting the superclass object

	// sigh...
	delete this;

	return true;

} // Release




//-----------------------------------------------------------------------------ReAllocate----------
// Herkules
// throw away all existing data and start from scratch 
//

bool 
CStoredObjectLibrary::ReAllocate()
{

	if( super::ReAllocate() )
	{

		// set the file offset for the data to 0
		// this will cause the next flush to allocate a new portion in the file
		m_MyFileOffsetData = 0;

		return true;
	}


	return false;

} // ReAllocate








#pragma pack( push, LibraryReference, 2 )

typedef struct 
{
	// the ID of the class the object belongs to
	STOREDCLASS_ID m_ClassID;

	// the index of the object 
	int	m_ObjectIndex;

	// the position in the file
	FILEOFFSET m_FileOffset;

} REFERENCEDATALIBRARY;

#pragma pack( pop, LibraryReference )



int 
CStoredObjectLibrary::SizeOfReference()
{
	return sizeof( REFERENCEDATALIBRARY );
}


unsigned char* 
CStoredObjectLibrary::CompileReference( unsigned char* _pData, IStoredObject* _pObj )
{
	REFERENCEDATALIBRARY Ref;

	if( _pObj )
	{
		Ref.m_ClassID = _pObj->GetClass()->GetID();
		Ref.m_ObjectIndex = _pObj->GetIndex();
		Ref.m_FileOffset = static_cast<CStoredObjectLibrary*>(_pObj)->m_MyFileOffsetHeader;
	}
	else
	{
		// means: invalid
		Ref.m_ClassID = -1;
		Ref.m_ObjectIndex = -1;
		Ref.m_FileOffset = 0;
	}

	*((REFERENCEDATALIBRARY*)_pData) = Ref;
	_pData += SizeOfReference();

	return _pData;

}

IStoredObject* 
CStoredObjectLibrary::CreateFromReferenceData( void* _pRefDat )
{

	REFERENCEDATALIBRARY* pRef = (REFERENCEDATALIBRARY*)_pRefDat; 
	
	// check for a NULL reference: FileOffset is 0
	if( pRef->m_FileOffset == 0 )
		return NULL;


	// retrieve a pointer to the class from the ID
	CStoredClassLibrary* pClass = static_cast<CStoredClassLibrary*>(GetClass()->GetStorage()->GetClass( pRef->m_ClassID ));

	// if class cannot be found, this is an error condition!
	assert( pClass );

	// if the class is not present, return...
	if( ! pClass )
		return NULL;


	// everything seems ok, so we can create an object and return the pointer
	return new CStoredObjectLibrary( pClass, pRef->m_ObjectIndex, pRef->m_FileOffset );

} // CreateFromReferenceData











// ----------------------------------------------------------------------------
//
//
// CStoredClassLibrary
//
//
// ----------------------------------------------------------------------------



// ----------------------------------------------------------------------------
// Constructor
CStoredClassLibrary::CStoredClassLibrary(	const string& _Classname, 
											const STOREDCLASS_ID _ClassID, 
											CStorage* _pStorage, 
											IStoredClass* const _pBaseClass ) :
super( _Classname, _ClassID, _pStorage, _pBaseClass ),
m_OffsetFirstObject( 0 ),
m_OffsetLastObject( 0 ),
m_NumObjects( 0 )
{
} 






// --------------------------------------------------------------------------------------------------
// Herkules
// add a new object 
//

IStoredObject* 
CStoredClassLibrary::AddObject( )
{


	//
	// create a new memory object.
	// the current counter is the index of that object
	//
	CStoredObjectLibrary* pO = new CStoredObjectLibrary( this );


	//
	// Note: The number of object available in the class is NOT increased 
	// here, bc. a library object is worth nothing as long as it isnt 
	// flushed out
	//

	return pO;

} // AddObject





// --------------------------------------------------------------------------------------------------
// Herkules
// add a new object 
//

CStoredVar* 
CStoredClassLibrary::CreateVar( const string& _Varname, STOREDVAR_TYPE _Type, STOREDVAR_ID _ID )
{

	// this is a significant change in repository
	static_cast<CStorageLibrary*>(m_pStorage)->m_bRepositoryDirty = true;
	
	return super::CreateVar( _Varname, _Type, _ID );

} // CreateVar







// ----------------------------------------------------------------------------
// static
//

CStoredClassLibrary* 
CStoredClassLibrary::ConstructFromFile( HANDLE _hFile, CStorageLibrary& _Storage  )
{
	// locals 
	unsigned long BytesRead;
	int i;

	CStoredClassLibrary* pNew = NULL;
	char ClassNameBuffer[MAXCLASSNAMELEN];
	char VarNameBuffer[MAXVARNAMELEN];

	STOREDCLASS_ID ClassID;
	STOREDCLASS_ID BaseClassID;

	
	CLASSNAMELEN ClassNameLen = 0;
	VARNAMELEN VarNameLen = 0;
	STOREDVAR_ID VarCount = 0;
	
	STOREDVAR_ID VarID;
	STOREDVAR_TYPE VarType;

	IStoredClass* pBaseClass = NULL;


	// first read the name for the class
	if( ! ReadFile( _hFile, &ClassNameLen, sizeof( CLASSNAMELEN ), &BytesRead, NULL ) )
		return NULL;
	

	// did we reach end of file?
	// this can happen if the last class in the file has already been read
	// this condition here is used as an indicator the loading has to be finished
	if( BytesRead == 0 )
		return NULL;


	// read the name of the class
	if( ! ReadFile( _hFile, &ClassNameBuffer, ClassNameLen, &BytesRead, NULL ) )
		return NULL;


	// add trailing 0
	ClassNameBuffer[ClassNameLen] = '\0';

	// read the ID of the class
	if( ! ReadFile( _hFile, &ClassID, sizeof( STOREDCLASS_ID ), &BytesRead, NULL ) )
		return NULL;


	//
	// Read the ID of an optional base class. It will be -1 if there is no base class
	//
	if( ! ReadFile( _hFile, &BaseClassID, sizeof( STOREDCLASS_ID ), &BytesRead, NULL ) )
		return NULL;
	
	if( BaseClassID != (STOREDCLASS_ID)-1 ) 
	{

		// Because base classes are always defined before the derived ones, it is
		// pretty sure that the derived class already exists and can be queried for.
		pBaseClass = _Storage.GetClass( BaseClassID );
	}


	//
	// create a new instance of the class
	//
	pNew = new CStoredClassLibrary( ClassNameBuffer, ClassID, &_Storage, pBaseClass );


	
	// number of objects
	if( ! ReadFile( _hFile, &pNew->m_NumObjects, sizeof( int ), &BytesRead, NULL ) )
		goto errorexit;


	// first  offset
	if( ! ReadFile( _hFile, &pNew->m_OffsetFirstObject, sizeof( FILEOFFSET ), &BytesRead, NULL ) )
		goto errorexit;


	// last offset 
	if( ! ReadFile( _hFile, &pNew->m_OffsetLastObject, sizeof( FILEOFFSET ), &BytesRead, NULL ) )
		goto errorexit;


	// number of variables
	if( ! ReadFile( _hFile, &VarCount, sizeof( STOREDVAR_ID ), &BytesRead, NULL ) )
		goto errorexit;

	
	//
	// create the variables
	//
	for( i = 0; i < VarCount; i++ )
	{

		// ID of variable
		if( ! ReadFile( _hFile, &VarID, sizeof( STOREDVAR_ID ), &BytesRead, NULL ) )
			goto errorexit;
		
		// Type of variable
		if( ! ReadFile( _hFile, &VarType, sizeof( STOREDVAR_TYPE ), &BytesRead, NULL ) )
			goto errorexit;

		
		//
		// read the name for the var
		//
		
		// length of name 
		if( ! ReadFile( _hFile, &VarNameLen, sizeof( VARNAMELEN ), &BytesRead, NULL ) )
			goto errorexit;


		// the name itself
		if( ! ReadFile( _hFile, &VarNameBuffer, VarNameLen, &BytesRead, NULL ) )
			goto errorexit;

		// add trailing 0
		VarNameBuffer[VarNameLen] = '\0';

		
		// now create it
		// we use the function from the super-class bc. ours also
		// marks repository as dirty
		CStoredVar* pNewVar = pNew->super::CreateVar( VarNameBuffer, VarType, VarID );

		// add the new var to the repository
		pNew->AddVarToRepository( pNewVar ); 

	}


	//
	// everything went OK
	//
	return pNew;


errorexit:
	delete pNew;
	return NULL;

}





// ----------------------------------------------------------------------------
bool
CStoredClassLibrary::DumpToFile( HANDLE _hFile )
{

	IStoredVarList* pVars = GetVarList();
	CStoredVar* pVar;
	unsigned long BytesWritten;

	CLASSNAMELEN ClassNameLen = 0;
	VARNAMELEN VarNameLen = 0;
	STOREDVAR_ID VarCount;

	ClassNameLen = (CLASSNAMELEN) m_Classname.size();

	if( ! WriteFile( _hFile, &ClassNameLen, sizeof( CLASSNAMELEN ), &BytesWritten, NULL ) )
		return false;

	if( ! WriteFile( _hFile, m_Classname.c_str(), ClassNameLen, &BytesWritten, NULL ) )
		return false;

	if( ! WriteFile( _hFile, &m_ClassID, sizeof( STOREDCLASS_ID ), &BytesWritten, NULL ) )
		return false;

	//
	// Write the ID of the baseclass or a -1, if base class is not present 
	//
	STOREDCLASS_ID BaseClassID = -1;

	if( GetBaseClass() )
		BaseClassID = GetBaseClass()->GetID();

	if( ! WriteFile( _hFile, &BaseClassID, sizeof( STOREDCLASS_ID ), &BytesWritten, NULL ) )
		return false;


	// first and last offset
	if( ! WriteFile( _hFile, &m_NumObjects, sizeof( int ), &BytesWritten, NULL ) )
		return false;


	// first and last offset
	if( ! WriteFile( _hFile, &m_OffsetFirstObject, sizeof( FILEOFFSET ), &BytesWritten, NULL ) )
		return false;

	if( ! WriteFile( _hFile, &m_OffsetLastObject, sizeof( FILEOFFSET ), &BytesWritten, NULL ) )
		return false;



	// number of variables
	VarCount = (STOREDVAR_ID)pVars->Count();
	if( ! WriteFile( _hFile, &VarCount, sizeof( STOREDVAR_ID ), &BytesWritten, NULL ) )
		return false;
	
	// now the variables
	pVar = (CStoredVar*)pVars->First();
	while ( pVar )
	{
		
		if( ! WriteFile( _hFile, &pVar->GetID(), sizeof( STOREDVAR_ID ), &BytesWritten, NULL ) )
			return false;
		
		if( ! WriteFile( _hFile, &pVar->GetType(), sizeof( STOREDVAR_TYPE ), &BytesWritten, NULL ) )
			return false;

		VarNameLen = (VARNAMELEN)pVar->Name().size();
		if( ! WriteFile( _hFile, &VarNameLen, sizeof( VARNAMELEN ), &BytesWritten, NULL ) )
			return false;

		if( ! WriteFile( _hFile, pVar->Name().c_str(), VarNameLen, &BytesWritten, NULL ) )
			return false;

		pVar = (CStoredVar*)pVars->Next();
	}


	//
	// release the variable collection again
	//
	pVars->Release();


	// everything is fine
	return true;


} // DumpToFile





//-----------------------------------------------------------------------------AssignObject--------
// Herkules
// Protected: Assign a new object at fileposition _Pos to the class. 
// Maintain the list of objects.
//

int
CStoredClassLibrary::AssignObject( FILEOFFSET _Pos ) 
{

	const HANDLE File = static_cast<CStorageLibrary*>(m_pStorage)->GetDatafile();
	DWORD BytesWritten;
	BOOL Ret;

	
	//
	// update links
	//
	
	//
	// only if its not the first object!
	//
	if( m_NumObjects )
	{

		//
		// move filepointer to the (previously) last object
		//
		DWORD Pos = SetFilePointer( File, m_OffsetLastObject, NULL, FILE_BEGIN );
		if( Pos != m_OffsetLastObject )
			return -1;


		//
		// write the previous EOF value there as a NEXT pointer
		//
		Ret = WriteFile( File, (LPCVOID)&_Pos, sizeof( FILEOFFSET ), &BytesWritten, NULL );
		if( ! Ret || BytesWritten != sizeof( FILEOFFSET ) ) 
			return -1;
	}
	else
	{
		// if it IS the frist object, also the appropriate offset
		m_OffsetFirstObject = _Pos;
	}


	// update the class
	m_OffsetLastObject = _Pos;
	
	
	// Return the current number of object as the new index and increase that number
	return m_NumObjects++;


} // AssignObject


//-----------------------------------------------------------------------------CreateHeaderInFile--
// Herkules
// Create an objects header in the data file
//

FILEOFFSET 
CStoredClassLibrary::CreateHeaderInFile( int& _ObjectIndex )
{

	const HANDLE File = static_cast<CStorageLibrary*>(m_pStorage)->GetDatafile();
	DWORD BytesWritten;
	BOOL Ret;

	_FILEOBJECTHEADER Header;

	// get position for new header in the file 
	DWORD PosNewHeader = static_cast<CStorageLibrary*>(m_pStorage)->EndOfObjects();


	// put the file-write pointer to the end of file
	DWORD EOFPos = SetFilePointer( File, PosNewHeader, NULL, FILE_BEGIN );


	// this has to be successfull
	if( EOFPos != PosNewHeader ) 
		return 0;

	//
	// setup the header structure
	//

	// currently our header is the end of the list
	// (will be patched later in file with next AppendObjectData())
	Header.m_Next = 0;					

	// Size and dataposition will be set when the data is written
	Header.m_Size = 0;
	Header.m_DataBlock = 0;

	
	//
	// write header to the file
	//
	Ret = WriteFile( File, (LPCVOID)&Header, sizeof( _FILEOBJECTHEADER ), &BytesWritten, NULL );
	if( ! Ret || BytesWritten != sizeof( _FILEOBJECTHEADER ) ) 
		return 0;

/*
	//
	// update links
	//
	
	//
	// only if its not the first object!
	//
	if( m_NumObjects )
	{

		//
		// move filepointer to the (previously) last object
		//
		DWORD Pos = SetFilePointer( File, m_OffsetLastObject, NULL, FILE_BEGIN );
		if( Pos != m_OffsetLastObject )
			return 0;


		//
		// write the previous EOF value there as a NEXT pointer
		//
		Ret = WriteFile( File, (LPCVOID)&EOFPos, sizeof( FILEOFFSET ), &BytesWritten, NULL );
		if( ! Ret || BytesWritten != sizeof( FILEOFFSET ) ) 
			return 0;
	}
	else
	{
		// if it IS the frist object, also the appropriate offset
		m_OffsetFirstObject = EOFPos;
	}


	// update the class
	m_OffsetLastObject = EOFPos;
	
	
	// the old number of objects is the index for the new one
	// hold on: this is a sacret moment, where an object gains its identity!
	_ObjectIndex = m_NumObjects;


	// increace number of availibele object
	m_NumObjects++;

*/

	_ObjectIndex = AssignObject( EOFPos );


	// this is a significant change in repository
	static_cast<CStorageLibrary*>(m_pStorage)->m_bRepositoryDirty = true;


	// update the storage object: where to put next object
	static_cast<CStorageLibrary*>(m_pStorage)->EndOfObjects() = EOFPos + sizeof( _FILEOBJECTHEADER );


	// return the old EOF position as a value for a reference
	// you will find the newly added object there
	return EOFPos; 


}





//-----------------------------------------------------------------------------AppendObjectData----
// Herkules
// Append a bunch of object data to the end of the file and maintain file-internal 
// offset-pointers
//

FILEOFFSET 
CStoredClassLibrary::AppendObjectData( char* _pData, OBJECTSIZE _FlushedSize, int& _ObjectIndex )
{
	const HANDLE File = static_cast<CStorageLibrary*>(m_pStorage)->GetDatafile();
	DWORD BytesWritten;
	BOOL Ret;

	_FILEOBJECTHEADER Header;


	// get position for new object in the file 
	DWORD PosNewObject = static_cast<CStorageLibrary*>(m_pStorage)->EndOfObjects();


	// put the file-write pointer to the end of file
	DWORD EOFPos = SetFilePointer( File, PosNewObject, NULL, FILE_BEGIN );


	// this has to be successfull
	if( EOFPos != PosNewObject ) 
		return NULL;


	// setup the header structure
	Header.m_Next = 0;					// will be patched in file with next AppendObjectData()
	Header.m_Size = _FlushedSize;


	// write the file with the data
	Ret = WriteFile( File, (LPCVOID)&Header, sizeof( _FILEOBJECTHEADER ), &BytesWritten, NULL );
	if( ! Ret || BytesWritten != sizeof( _FILEOBJECTHEADER ) ) 
		return 0;

	Ret = WriteFile( File, (LPCVOID)_pData, _FlushedSize, &BytesWritten, NULL );
	if( ! Ret || BytesWritten != _FlushedSize ) 
		return 0;


	//
	// update links
	//
	
	//
	// only if its not the first object!
	//
	if( m_NumObjects )
	{

		//
		// move filepointer to the (previously) last object
		//
		DWORD Pos = SetFilePointer( File, m_OffsetLastObject, NULL, FILE_BEGIN );
		if( Pos != m_OffsetLastObject )
			return 0;


		//
		// write the previous EOF value there as a NEXT pointer
		//
		Ret = WriteFile( File, (LPCVOID)&EOFPos, sizeof( FILEOFFSET ), &BytesWritten, NULL );
		if( ! Ret || BytesWritten != sizeof( FILEOFFSET ) ) 
			return 0;
	}
	else
	{
		// if it IS the frist object, also the appropriate offset
		m_OffsetFirstObject = EOFPos;
	}


	// update the class
	m_OffsetLastObject = EOFPos;
	
	


	// the old number of objects is the index for the new one
	_ObjectIndex = m_NumObjects;


	// increace number of availibele object
	m_NumObjects++;


	// this is a significant change in repository
	static_cast<CStorageLibrary*>(m_pStorage)->m_bRepositoryDirty = true;


	// update the storage object: where to put next object
	static_cast<CStorageLibrary*>(m_pStorage)->EndOfObjects() = EOFPos + sizeof( _FILEOBJECTHEADER ) + _FlushedSize;


	// return the old EOF position as a value for a reference
	// you will find the newly added object there
	return EOFPos; 

} // AppendObjectData





//-----------------------------------------------------------------------------UpdateObjectData----
// Herkules
// Update objects data in the file
//

bool
CStoredClassLibrary::UpdateObjectData( FILEOFFSET _Where, char* _pData, OBJECTSIZE _FlushedSize )
{
	// locals
	const HANDLE File = static_cast<CStorageLibrary*>(m_pStorage)->GetDatafile();
	DWORD BytesWritten;
	BOOL Ret;
	DWORD Pos;


	// put the file-write pointer to the position of the  object
	// thereby skip the objects header
	Pos = SetFilePointer( File, _Where + sizeof( _FILEOBJECTHEADER ) , NULL, FILE_BEGIN );
	if( Pos != _Where + sizeof( _FILEOBJECTHEADER ) )
		return FALSE;


	// write the file with the data
	Ret = WriteFile( File, (LPCVOID)_pData, _FlushedSize, &BytesWritten, NULL );
	if( ! Ret || BytesWritten != _FlushedSize ) 
		return false;

	// everything went fine
	return true;

} // UpdateObjectData


/*bool
CStoredClassLibrary::UpdateObjectData( FILEOFFSET _Where, char* _pData, OBJECTSIZE _FlushedSize )
{
	// locals
	const HANDLE File = static_cast<CStorageLibrary*>(m_pStorage)->GetDatafile();
	DWORD BytesWritten;
	BOOL Ret;
	DWORD Pos;


	// put the file-write pointer to the position of the  object
	// thereby skip the objects header
	Pos = SetFilePointer( File, _Where + sizeof( _FILEOBJECTHEADER ) , NULL, FILE_BEGIN );
	if( Pos != _Where + sizeof( _FILEOBJECTHEADER ) )
		return FALSE;


	// write the file with the data
	Ret = WriteFile( File, (LPCVOID)_pData, _FlushedSize, &BytesWritten, NULL );
	if( ! Ret || BytesWritten != _FlushedSize ) 
		return FALSE;

	// everything went fine
	return TRUE;

} // UpdateObjectData
*/







//-----------------------------------------------------------------------------GetObjectList-------
// Herkules
// create a new list-style access object
//

IStoredObjectList* 
CStoredClassLibrary::GetObjectList()
{ 

	return new CStoredObjectListLibrary( *this, static_cast<CStorageLibrary*>(m_pStorage)->GetDatafile(), m_NumObjects ); 

} // GetObjectList




//-----------------------------------------------------------------------------GetObjectArray------
// Herkules
// create a new list-style access object
//

IStoredObjectArray* 
CStoredClassLibrary::GetObjectArray()
{ 

	return new CStoredObjectArrayLibrary( *this, static_cast<CStorageLibrary*>(m_pStorage)->GetDatafile() ); 

} // GetObjectArray







// --------------------------------------------------------------------------------
//
//
//
// CStoredObjectListLibrary 
//
//
//
// --------------------------------------------------------------------------------




//------------------------------------------------------------------------------------------------
// Constructor
//

CStoredObjectListLibrary::CStoredObjectListLibrary( CStoredClassLibrary& _Class, HANDLE _File, const int _Count ) :
m_Class( _Class ),
m_DatFile( _File ),
//m_NextOffset( _Class.GetFirstObjectOffset() ),
m_Count( _Count ),
m_IndexCounter( 0 )
{

	GotoFirst();
	ReadHeader();

}



//------------------------------------------------------------------------------------------------
CStoredObject* 
CStoredObjectListLibrary::GetCurrent( bool _WithData ) 
{

	// locals
	CStoredObjectLibrary* pObj;
//	char* pData;

	
	///////////
	// create an object
	pObj = new CStoredObjectLibrary( &m_Class, m_IndexCounter, m_CurrentOffset, m_CurrentHeader.m_Size );


	// allocate memory for the data
//	pData = pObj->AllocateCompiledObject( m_CurrentHeader.m_Size );


	// assign buffer  to object
//	pObj->AssignData( pData, m_CurrentHeader.m_Size );

	// load data only if _WithData is set
	// but if loading fails, delete object and return false
	if( _WithData )
		if( ! pObj->Refresh() )		goto error_exit;


	return pObj;

// jump here in case of error
error_exit:

	delete pObj;
	return NULL;


} // GetCurrent



//------------------------------------------------------------------------------------------------
// Herkules
//

bool 
CStoredObjectListLibrary::GotoFirst()
{

	// when we start a new iteration, index counter has to be 0 of course
	m_IndexCounter = 0;


	// Offset of first object in file
	m_CurrentOffset = m_Class.GetFirstObjectOffset();

	return ( m_CurrentOffset != 0 ); 
}


//------------------------------------------------------------------------------------------------
bool 
CStoredObjectListLibrary::ReadHeader()
{

	// locals
	DWORD Pos;
	DWORD BytesRead;

	//
	// now do the file stuff
	//

	// put read pointer to begin of object 
	Pos = SetFilePointer( m_DatFile, m_CurrentOffset, NULL, FILE_BEGIN );

	// if not at start of file return
	if( Pos != m_CurrentOffset )	return false;
	

	// read in the header
	if( ! ReadFile( m_DatFile, &m_CurrentHeader, sizeof( _FILEOBJECTHEADER ), &BytesRead, NULL ) )	return false;


	return true;
}


//------------------------------------------------------------------------------------------------
bool 
CStoredObjectListLibrary::GotoNext()
{

	// update offset for the Next() method
	m_CurrentOffset = m_CurrentHeader.m_Next;
	m_IndexCounter++;

	return ( m_CurrentOffset != 0 ); 
}






//------------------------------------------------------------------------------------------------
// retrieve pointer to first object
//

IStoredObject*	
CStoredObjectListLibrary::First( bool _WithData )
{

	//
	// for a multithreaded environemt, this must be blocked against other file positioning operations
	//

	if( GotoFirst() )
	{
		ReadHeader();

		return GetCurrent( _WithData );
	}

	return NULL;

} // First







//------------------------------------------------------------------------------------------------
// retrieve pointer to Next object
//

IStoredObject*	
CStoredObjectListLibrary::Next( bool _WithData )
{

	//
	// for a multithreaded environemt, this must be blocked against other file positioning operations
	//


	// increment index counter 

	if( GotoNext() )
	{

		ReadHeader();

		return GetCurrent( _WithData );
	}

	return NULL;


} // Next






// --------------------------------------------------------------------------------
//
//
//
// CStoredObjectListLibrary 
//
//
//
// --------------------------------------------------------------------------------





//------------------------------------------------------------------------------------------------
// Constructor
//

CStoredObjectArrayLibrary::CStoredObjectArrayLibrary( CStoredClassLibrary& _Class, HANDLE _File ) :
m_LastIndex( 0 )
{
	
	m_pList = static_cast<CStoredObjectListLibrary*>(_Class.GetObjectList());

}




//------------------------------------------------------------------------------------------------
// Destructor
//

CStoredObjectArrayLibrary::~CStoredObjectArrayLibrary()
{

	m_pList->Release();

}




//------------------------------------------------------------------------------------------------
// Access the array
//

IStoredObject*
CStoredObjectArrayLibrary::Get( int _Index, bool _WithData )
{
	IStoredObject* pObj;

	// Index has to be in range
	if( _Index >= m_pList->Count() )
		return NULL;


	// the overall behaviour of this is implemented using the list
	// the array is completely virtual

	if( _Index == 0 )
	{
		pObj = m_pList->First( _WithData );
	}
	else if( _Index == m_LastIndex )
	{
		pObj = m_pList->GetCurrent( _WithData );
	}
	else if( _Index > m_LastIndex ) 
	{
		m_pList->GotoNext();
		for( int i = 0; i < _Index - m_LastIndex - 1; i++ )
		{
			m_pList->ReadHeader();
			m_pList->GotoNext();
		}
		m_pList->ReadHeader();
		pObj = m_pList->GetCurrent( _WithData );

	} 
	else 
	{
		if( m_pList->GotoFirst() ) 
		{
			m_pList->ReadHeader();
			m_LastIndex = 0;
			pObj = Get( _Index, _WithData );
		}
		else
			return NULL;
	}


	m_LastIndex = _Index;

	return pObj;

} // Get




// --------------------------------------------------------------------------------
//
//
//
// CStorageLibrary 
//
//
//
// --------------------------------------------------------------------------------






// --------------------------------------------------------------------------------
// Herkules
// Constructor
//

CStorageLibrary::CStorageLibrary( IStorageEvent* _pEvent ) :
super( _pEvent )
{
	
	// at construction time, everything is clean
	m_bRepositoryDirty = false;


} // CStorageLibrary




// --------------------------------------------------------------------------------
// Herkules
// Destructor
//

CStorageLibrary::~CStorageLibrary() 
{ 	
	
	// if the lib is still open, close it now.
	if( m_hDataFile != INVALID_HANDLE_VALUE )
		Close();

} // ~CStorageLibrary





// --------------------------------------------------------------------------------
// Herkules
// Create a Class
//

CStoredClass* 
CStorageLibrary::CreateClass( const string& _Classname, const STOREDCLASS_ID _ClassID, IStoredClass* const _pBaseClass )	
{ 

	// this is a significant change in repository
	m_bRepositoryDirty = true;
	

	return new CStoredClassLibrary( _Classname, _ClassID, this, _pBaseClass ); 

} // CreateClass






// --------------------------------------------------------------------------------
// Herkules
// Connect with the lib
//

bool 
CStorageLibrary::Connect( const char* _Connection )
{

	super::Connect( _Connection );	
	
	//
	// remember pathname where to find/create files
	//
	strcpy( m_LibPath, _Connection );



	//
	// now the data file
	//
	if( ! OpenDatafile() )
	{
		// close file again
		CLOSEHANDLE( m_hDataFile );
			
		return false;
	}

	// everything is fine
	return true;
}




// --------------------------------------------------------------------------------
// Herkules
//

bool 
CStorageLibrary::WriteDatafileHeader()
{

	//
	// set the filepointer to begin of file
	//

	DWORD Pos;

	// put read pointer to beginning of the file
	Pos = SetFilePointer( m_hDataFile, 0, NULL, FILE_BEGIN );

	// if not at start of file, return
	if( Pos != 0)
		return false;



	//
	// write the header
	//

	_LIBRARYFILEHEADER FileHeader;
	unsigned long BytesWritten = 0;


	FileHeader.m_Magic = LIBDATAMAGIC;
	FileHeader.m_EndOfObjects = m_EndOfObjects;


	// file must be writable
	if( ! WriteFile( m_hDataFile, &FileHeader, sizeof( _LIBRARYFILEHEADER ), &BytesWritten, NULL ) )
		return false;

	// file must be as least sizeof(int) bytes long now
	if( BytesWritten != sizeof( _LIBRARYFILEHEADER ) )
		return false;


	return true;

} // WriteDatafileHeader





// --------------------------------------------------------------------------------
// Herkules
// Close the lib
//

bool 
CStorageLibrary::Close( )
{

	//
	// save index to file
	//
	if( ! SaveIndex() )
		return false;


	//
	// flush the position of repository
	// 
	WriteDatafileHeader();



	//
	// close the datafile
	//
	CLOSEHANDLE( m_hDataFile );
	

	// clean again
	m_bRepositoryDirty = false;


	return true;

} // Close







// --------------------------------------------------------------------------------
// Herkules
// just check for the magic number
//

bool 
CStorageLibrary::IsIndexfile()
{
 
	unsigned int Magic = 0;
	unsigned long BytesRead = 0;
	DWORD Pos;

	// put read pointer to beginning of the repository 
	Pos = SetFilePointer( m_hDataFile, m_EndOfObjects, NULL, FILE_BEGIN );


	// if not at start of repository, return
	// maybe file is too short?
	if( Pos != m_EndOfObjects )
		return false;


	// file must be readable
	if( ! ReadFile( m_hDataFile, &Magic, sizeof( int ), &BytesRead, NULL ) )
		return false;

	// file must be as least sizeof(int) bytes long
	if( BytesRead != sizeof( int ) )
		return false;

	// the first int in file has to be the magic number
	if( Magic != LIBINDEXMAGIC )
		return false;

	// passed all tests
	return true;
}




// --------------------------------------------------------------------------------
// Herkules
// just check for the magic number
//

bool 
CStorageLibrary::IsDatafile()
{
 
	// locals
	unsigned long BytesRead = 0;
	DWORD Pos;

	_LIBRARYFILEHEADER FileHeader;


	// put read pointer to begin of file 
	Pos = SetFilePointer( m_hDataFile, 0, NULL, FILE_BEGIN );

	// if not at start of file return
	if( Pos != 0 )
		return false;
	

	// file must be readable; read header
	if( ! ReadFile( m_hDataFile, &FileHeader, sizeof( _LIBRARYFILEHEADER ), &BytesRead, NULL ) )
		return false;


	// file must be as least sizeof(int) bytes long
	if( BytesRead != sizeof( _LIBRARYFILEHEADER ) )
		return false;


	// the first int in file has to be the magic number
	if( FileHeader.m_Magic != LIBDATAMAGIC )
		return false;


	//
	// option!
	// we could check wether FileHeader.m_EndOfObjects fits with size of file! 
	//

	// tests ok so far, get the position of the repository on the file
	m_EndOfObjects = FileHeader.m_EndOfObjects;
		

	// passed all tests
	return true;

}






// --------------------------------------------------------------------------------
// Herkules
//

bool 
CStorageLibrary::SaveIndex()
{

	// if nothing has changed, we can be very lazy...
	if( ! m_bRepositoryDirty )	
		return true;


	// data file has to be open
	assert( m_hDataFile );



	//
	// set the filepointer behind all objects
	//

	DWORD Pos;

	// put read pointer to beginning of the repository 
	Pos = SetFilePointer( m_hDataFile, m_EndOfObjects, NULL, FILE_BEGIN );

	// if not at start of repository, return
	// maybe file is too short?
	if( Pos != m_EndOfObjects )
		return false;



	//
	// write index header
	//

	unsigned int Magic = LIBINDEXMAGIC;
	unsigned long BytesWritten = 0;

	// file must be writable
	if( ! WriteFile( m_hDataFile, &Magic, sizeof( int ), &BytesWritten, NULL ) )
		return false;

	// file must be as least sizeof(int) bytes long now
	if( BytesWritten != sizeof( int ) )
		return false;



	//
	// write repository data
	//


	// get the list of all classes in repository
	IStoredClassList* pClasses = GetClassList();


	//
	// now loop on all classes and dump them to file
	//

	// get first class 
	CStoredClassLibrary* pClass = static_cast<CStoredClassLibrary*>(pClasses->First());


	// for all classes in the list 
	while( pClass )
	{

		// dump a single class
		if( ! pClass->DumpToFile( m_hDataFile ) )
		{
			// did not work?

			// release the classes list
			pClasses->Release();

			// flag failure
			return false;

		}

		// get next class
		pClass = static_cast<CStoredClassLibrary*>(pClasses->Next());

	}


	// free class iterator
	pClasses->Release();

	
	// fine!
	return true;

} // SaveIndex





//-----------------------------------------------------------------------------------------------------------
// Herkules
// Load the index from the file to repository
//

bool 
CStorageLibrary::LoadIndex()
{
	// locals
	CStoredClassLibrary* pClass;

	// data file must be open
	assert( m_hDataFile );


	// check wether it IS an index in the file 
	if ( ! IsIndexfile() )
		return false;


	// as long as the file has more classes....
	while( pClass = CStoredClassLibrary::ConstructFromFile( m_hDataFile, *this ) )
	{

		// add the class to repository
		AddClassToRepository( pClass );

	}

	// no problem
	return true;

} // LoadIndex





// --------------------------------------------------------------------------------
// Herkules
//

bool 
CStorageLibrary::OpenDatafile()
{

	unsigned int Magic = LIBDATAMAGIC;
	unsigned long BytesWritten = 0;


	// 
	// open data file for reading/writing
	//
	m_hDataFile = CreateFile(
						m_LibPath,
				//ConstructFilename( LibData ),
						GENERIC_WRITE | GENERIC_READ,
						FILE_SHARE_READ,
						NULL,
						OPEN_EXISTING,
						FILE_ATTRIBUTE_NORMAL,
						NULL );


	// if the file does not exist
	if( m_hDataFile == INVALID_HANDLE_VALUE )
	{

		// 
		// create new data file for reading/writing
		//
		m_hDataFile = CreateFile(
							m_LibPath,
				//ConstructFilename( LibData ),
							GENERIC_WRITE | GENERIC_READ,
							FILE_SHARE_READ,
							NULL,
							OPEN_ALWAYS,
							FILE_ATTRIBUTE_NORMAL,
							NULL );

		// oh oh....
		if( m_hDataFile == INVALID_HANDLE_VALUE )
			return false;

		//
		// prepare the new library file
		//
		m_EndOfObjects = sizeof( _LIBRARYFILEHEADER );
		if( ! WriteDatafileHeader() )
			return false;

	} 
	else
	{


		//
		// file is open and initialized now
		//
		
		
		// if there is a file and it is not an data file
		// this IS an error
		if( ! IsDatafile() )	
			return false;


		if( ! LoadIndex() )
			return false;
	
	}

	// everything went fine
	return true;

} // OpenDatafile




