// CStorageLibrary.h
//**********************************************************************************************
//
//
//				(C) Copyright 1998 by Dipl. Phys. Joerg Plewe, HARDCODE Development
//					All rights reserved. Copying, modification,
//					distribution or publication without the prior written 
//					consent of the author is prohibited.
//
//
//	Filename:		CStorageLibrary.h
//	Author:			Herkules
//  Created:		9810xx 
//
//	Module:			StorageLibrary
//  Dependecies:	CStorageMemory.h/.cpp
//
//  History:
//  $Date$		$Who$		$What$
//	19981221	Herkules	Integrated new iterator semantic. See changes in CStorage.h 
//	19981228	Herkules	BUG: When Flush()ing objects with _KeepData==false more than once, 
//							it re-Compiles() and saves an already cleared object!
//							Fix by checking the returncode of super::Flush() when doing our own. 
//	19990203	Herkules	Introduced CStoredObject::ReAllocate()
//
/*	Purpose:

These are the library 'driver' classes for the storage engine.
Classes are derived from the classes in CStorage.h and CStorageMemory.h

*/
//**********************************************************************************************

#pragma once


#ifndef _H_CStorageLibrary
#define _H_CStorageLibrary


#include "../Memory/CStorageMemory.h" 


//
// defines
//

#define LIBINDEXMAGIC	(*(int*)"ules")
#define LIBDATAMAGIC	(*(int*)"Herk")


//
// typedefs
//
typedef unsigned int FILEOFFSET;


//-----------------------------------------------------------------------------_LIBRARYFILEHEADER
// Herkules
// Header of a library file
//

typedef struct 
{
	// Magic
	int m_Magic;

	// points behind last object, begin of repository
	FILEOFFSET m_EndOfObjects;

} _LIBRARYFILEHEADER;



//-----------------------------------------------------------------------------_FILEOBJECTHEADER---
// Herkules
// Header for each object in the file
//

typedef struct 
{
	
	// fileposition of next header
	// MUST be the first element in this structure. AppendObjectData() patches it without
	// asking for the structure of the header
	FILEOFFSET m_Next;

	// compiled size of the object, needed for memory allocation when loading
	OBJECTSIZE m_Size;

	// Link to the real data block.
	// Normally it followes directly after this header, but ReAllocate() may force 
	// the block to a new position
	FILEOFFSET m_DataBlock;

} _FILEOBJECTHEADER;




//
// forwards
//
class CStorageLibrary;
class CStoredClassLibrary;
class CStoredObjectListLibrary;










//----------------------------------------------------------------------------------------
// the object in the library
//

class CStoredObjectLibrary : public CStoredObjectMemory
{
	typedef CStoredObjectMemory super;


public:


	virtual bool Flush( bool _KeepData = false );
	virtual bool Refresh();
	virtual bool Release();


	// throw away all existing data and start from scratch 
	virtual bool ReAllocate();


protected:

	// only the class  and the iterators are allowed to create the object
	friend class CStoredClassLibrary;
	friend class CStoredObjectListLibrary;
	friend class CStoredObjectArrayLibrary;

	

	// constructor
	CStoredObjectLibrary( CStoredClassLibrary* _pStoredClass );

	// Constructor to be used by the iterators or reference resolution
	// If the _OffsetData and _CompiledSize are not given, they have to be
	// read from the header at _OffsetHeader
	CStoredObjectLibrary( CStoredClassLibrary* _pStoredClass, int _Index, 
							FILEOFFSET _OffsetHeader, 
							FILEOFFSET _OffsetData = 0, 
							OBJECTSIZE _CompiledSize = 0 );


	// overloads for reference handling
	virtual int SizeOfReference();
	virtual unsigned char* CompileReference( unsigned char* _pData, IStoredObject* _pObj );
	virtual IStoredObject* CreateFromReferenceData( void* _pRefDat );

	
	// 
	// handling the file
	// 
	bool SaveDataToFile();


private:

	// offset in file where the object header can be found
	// this may beused for storage of references later
	// = 0: not assigned
	FILEOFFSET m_MyFileOffsetHeader;

	// where will the objects data be in the file? This information is part 
	// of the header, which can be found at m_MyFileOffsetHeader
	// = 0: not assigned
	FILEOFFSET m_MyFileOffsetData;


}; // CStoredObjectLibrary










// ----------------------------------------------------------------------------
// now derived from CStoredClassMemory
// deriving from CStoredClass should be sufficient
//

class CStoredClassLibrary : public CStoredClassMemory
{
	typedef CStoredClassMemory super;


public:
	// Constructor
	CStoredClassLibrary(	const string& _Classname, 
							const STOREDCLASS_ID _ClassID, 
							CStorage* _pStorage,
							IStoredClass* const _pBaseClass = NULL );

	// Destructor
	virtual					~CStoredClassLibrary() {}


	//
	// the interface

	// add data (inherited function)
	virtual IStoredObject*	AddObject( );


	//
	// get access objects
	//
	virtual IStoredObjectArray* GetObjectArray();
	virtual IStoredObjectList* GetObjectList();




	// 
	// other functions
	//

	// dump the class to a file while saving repository
	bool DumpToFile( HANDLE _hFile );

	// create all variables from a file
	static CStoredClassLibrary* ConstructFromFile( HANDLE _hFile, CStorageLibrary& _Storage );


	// append object data to file 
	FILEOFFSET AppendObjectData( char* _pData, OBJECTSIZE _FlushedSize, int& _ObjectIndex );

	// update a bunch of data somewhere in the file
	bool UpdateObjectData( FILEOFFSET _Where, char* _pData, OBJECTSIZE _FlushedSize );



	// Create a new header in the datafile
	FILEOFFSET CreateHeaderInFile( /*[out]*/ int& _ObjectIndex );




	FILEOFFSET GetFirstObjectOffset()	const { return m_OffsetFirstObject; }

protected:

	// the CreateVar method changes the repository. So we must set the flag!
	virtual CStoredVar* CreateVar( const string& _Varname, STOREDVAR_TYPE _Type, STOREDVAR_ID _ID );

	// Assign a new object a fileposition _Pos to this class. Maintain the persistent list of objects
	int CStoredClassLibrary::AssignObject( FILEOFFSET _Pos );



	// number of objects in class
	int m_NumObjects;

	// configuration data for the storer
	FILEOFFSET m_OffsetFirstObject;
	FILEOFFSET m_OffsetLastObject;
	


private:

//	CStorageLibrary& m_Storage;

}; // CStoredClassLibrary










// ----------------------------------------------------------------------------
//
// class for object access in an list manner
//

class CStoredObjectListLibrary : public CStoredObjectList
{
	friend class CStoredClassLibrary;
public:

	// number of objects
	virtual int	Count()		 	{ return m_Count; }

	// retrieve pointer to first object
	virtual IStoredObject* First( bool _WithData );

	// pointer to next object
	virtual IStoredObject* Next( bool _WithData );


protected:
	CStoredObjectListLibrary( CStoredClassLibrary& _Class, HANDLE _File, const int _Count );


private:

	//
	// variables
	//

	CStoredClassLibrary& m_Class;

	HANDLE m_DatFile;
	const int m_Count;
	FILEOFFSET m_NextOffset;

	int m_IndexCounter;



	// 
	// extent the list with a couple of functions making it usable by 
	// CStoredObjectArrayLibrary	
	//
	friend class CStoredObjectArrayLibrary;

	_FILEOBJECTHEADER m_CurrentHeader;
	FILEOFFSET m_CurrentOffset;

	CStoredObject* GetCurrent( bool _WithData );
	bool GotoFirst();
	bool GotoNext();
	bool ReadHeader();



};








// ----------------------------------------------------------------------------
//
// class for object access in an list manner
//

class CStoredObjectArrayLibrary : public CStoredObjectArray
{
	friend class CStoredClassLibrary;
public:

	// number of objects
	virtual int	Count()		 	{ return m_pList->Count(); }

	// retrieve pointer to first object
	virtual IStoredObject*	Get( int _Index, bool _WithData );



protected:
	CStoredObjectArrayLibrary( CStoredClassLibrary& _Class, HANDLE _File );

	~CStoredObjectArrayLibrary();
private:

	//
	// variables
	//

	// the array is complete virtual and relies on a list
	CStoredObjectListLibrary* m_pList;

	int m_LastIndex;
};








// ----------------------------------------------------------------------------
// the storage class for libraries
//

class CStorageLibrary : public CStorage
{
	typedef CStorage super;

public:

	CStorageLibrary( IStorageEvent* _pEvent = NULL );

	// overloaded functions
	virtual bool Connect( const char* _Connection );
	virtual bool Close();							// save index and close all files


	const HANDLE GetDatafile()		{ return m_hDataFile; }

	bool m_bRepositoryDirty;


	FILEOFFSET& EndOfObjects()		{ return m_EndOfObjects; }


protected:

	virtual ~CStorageLibrary();

	//
	// 
	//
	virtual CStoredClass* CreateClass(	const string& _Classname, 
										const STOREDCLASS_ID _ClassID, 
										IStoredClass* const _pBaseClass = NULL );


private:

	bool IsIndexfile();				
	bool IsDatafile();
	bool SaveIndex();
	bool LoadIndex();
	bool OpenDatafile();
	bool WriteDatafileHeader();


	//
	// variables
	//
	
	// file- and pathnames
	char m_LibPath[MAX_PATH];
	char m_FilenameBuffer[MAX_PATH];

	// handles for Windows objects
	HANDLE m_hDataFile;


	// Offset were next new object will go
	// or: where index will be stored on Close()
	FILEOFFSET m_EndOfObjects;

};




#endif // _H_CStorageLibrary





/* Library format

Indexfile:
Magic
*IndexEntry:
	SizeofRecord[short]				// recordlength of the index entry
	Classname[countedstring]
	OffsetFirstObject[int]
	OffsetLastObject[int]
	NumberOfVars[short]
	*Variable:
		VariableID[short]
		VariableType[short]
		VariableName[countedstring]


Datafile:
Magic
*Dataset
	Offset next object of same class
	SizeOfObject
	VariableID[short]
	Value[variant]
*/