// CStorageMemory.cpp
//**********************************************************************************************
//
//
//				(C) Copyright 1998 by Dipl. Phys. Joerg Plewe, HARDCODE Development
//					All rights reserved. Copying, modification,
//					distribution or publication without the prior written 
//					consent of the author is prohibited.
//
//
//	Filename:		CStorageMemory.cpp
//	Author:			Herkules
//  Created:		9810xx 
//
//	Module:			StorageMemory
//  Dependecies:	CStorage classes
//
//  History:
//  $Date$		$Who$		$What$
//	19981220	Herkules	GetValue() now receives a <type>*& instead of the pointer value
//							GetValue() now checks the type of pointer against the type of the 
//							variable in case of a NULL pointer. Only of the types fit, immediate
//							pointer assignment makes sense.
//							The access templates have been changed accordingly 
//	19981223	Herkules	Added datatype LongInt, string conversion is not yet OK
//	19990203	Herkules	Implemented CStoredObject::ReAllocate()
//
/*	Purpose:

Implementation of the memory storage classes.

*/
//**********************************************************************************************


#define EXPORT_STORAGE


#include "CStorageMemory.h"
#include <assert.h>

// we need string stream for the conversion
#include <sstream>
#include <iomanip>


typedef unsigned int VALUESKIPDISTANCE;
typedef unsigned int COMPILEDVALUEOFFSET;




//-----------------------------------------------------------------------------REFERENCEDATAMEMORY-
// Herkules
// flat data structure describing a reference in a memory database
//

typedef struct 
{
	IStoredObject* m_pObject;

} REFERENCEDATAMEMORY;




//-------------------------------------------------------------------------------------------------
//
// CStoredObjectMemory                                                         CStoredObjectMemory
//
//-------------------------------------------------------------------------------------------------



//-----------------------------------------------------------------------------~CStoredObjectMemory
// Destructor
//

CStoredObjectMemory::~CStoredObjectMemory()
{
	if( m_pCompiledObject )
	{
		DeAllocateCompiledObject();
	}
	else
	{
		
		std::list<IUncompiledValue*>::iterator It;

		for( It = m_lUncompiledValues.begin(); It != m_lUncompiledValues.end(); ++It )
		{
			delete (*It);
		}
		m_lUncompiledValues.clear();
	}

} // Destructor









/////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
// utilities
//
//
/////////////////////////////////////////////////////////////////////////////////////////////////////


//-------------------------------------------------------------------------------------------------
// little macro that makes sure that only compiled objects are accessed
//
#define COMPILEDONLY	if( ! m_pCompiledObject ) return false;
#define UNCOMPILEDONLY	if(   m_pCompiledObject ) return false;





//-----------------------------------------------------------------------------_CompiledHeaderDirectoryEntry
// Herkules
// The memory storage uses a compiled format consisting if a 
// directory area and a data area.
// The directory as an array of the following structure:
//
#pragma pack( push, CompiledDirectory, 2 )

typedef struct 
{
	//
	// the ID of the variable
	//
	STOREDVAR_ID ID;

	//
	// the offset from the actual directory entry to the start of the value 
	//
	COMPILEDVALUEOFFSET Offset;

} _CompiledHeaderDirectoryEntry;

#pragma pack( pop, CompiledDirectory )








//-----------------------------------------------------------------------------FindCompiledValue---
// Herkules
// utility function
// Find a value in an compiled object
//

static char* 
FindCompiledValue( CStoredVar* _pVar, unsigned char* _pDataBuffer )
{

	// pointer into data array
	unsigned char* pData = _pDataBuffer;

	//
	// retrieve the number of values in the compiled field
	//
	STOREDVAR_ID NumberOfValues = *(STOREDVAR_ID*)pData;		
	pData +=sizeof( STOREDVAR_ID );


	// pointer into directory
	_CompiledHeaderDirectoryEntry* pDir = (_CompiledHeaderDirectoryEntry*)pData;

	// for all values in directory
	for( STOREDVAR_ID i = 0; i < NumberOfValues; i++ )
	{

		//
		// get the ID of the variable from the compiled field
		//
		STOREDVAR_ID VarID = pDir->ID;

		//
		// is it the variable we are looking for?
		//
		if( VarID == _pVar->GetID() )
		{
			// yeah, we found it!
			return (char*)pDir + pDir->Offset;
		} 

		// next directory entry
		pDir++;
	}

	// nothing found
	return NULL;

} // FindCompiledValue




//-----------------------------------------------------------------------------CopyTypedMemory-----
// Herkules
// utility function
// Copies a number of values with type srcT into an array of type destT
// !! this is NOT equivalent to CopyMemory!
//

template <class destT, class srcT> void
CopyTypedMemory( destT* _pDest, srcT* _pSrc, int srcCount )
{
	int i;

	for( i = 0; i < srcCount; i++ )
		_pDest[i] = static_cast<destT>(_pSrc[i]);

} // CopyTypedMemory<destT,srcT>()




//-----------------------------------------------------------------------------AssignSingleValueToArray
// Herkules
// utility template function
// Assigns a single (no-array-typed) value to an array
//
// 19981220	Herkules	_pValue is now a reference of a pointer. 
//

template <class srcT, class destT> 
bool
AssignSingleValueToArray( srcT* _pSrc, destT*& _pValue, int& _Size )
{

	// if someone retrieved immediate access to our internal data (e.g. with a 
	// NULL  pointer in GetValue()), _pData and _pValue may be identical
	// in this case, there is nothing to move at all
	if( (void*)_pSrc == (void*)_pValue )
		return true;


	// assign the value to the first element of _pValue
	//
	// we may do this ONLY if the user asks for a single element!
	// if he wants to get or set more, we fail and tell him (_Size = 1)
	// that there is only one available
	//
	if( _Size == 1 )
	{

		// check type of the pointer
		if( _pValue )
		{
			// _pValue != 0: assign the value
			*_pValue = static_cast<destT>(*_pSrc);
		}
		else
		{
			// _pValue == 0: assign the address
			_pValue = (destT*)_pSrc;
		}

		return true;
	}
	else
	{
		// if _Size == 0, we don't want to assign anything
		bool SizeIs0 = (_Size == 0);

		// there only is one value!
		_Size = 1;


		// if the size is 0 and _pValue == NULL, we need to assign the pointer, too
		if( SizeIs0 && ! _pValue )
			_pValue = (destT*)_pSrc;


		//
		// in case program asked for 0 values, this query is OK too
		// and we flag success.
		// otherwise it was a failure
		//
		return SizeIs0;
	}

} // AssignSingleValueToArray



//-------------------------------------------------------------------------------------------------
// Herkules
// utility template function
//
// 19981220	Herkules	return when _pData and _pValue are equal
//

template <class destT, class srcT> bool
SetTypedArray( destT* _pData, srcT* _pValue, int& _Size )
{ 

	
	// if someone retrieved immediate access to our internal data (e.g. with a 
	// NULL  pointer in GetValue()), _pData and _pValue may be identical
	// in this case, there is nothing to move at all
	if( (void*)_pData == (void*)_pValue )
		return true;
	

	// remember size to get
	int SetSize = _Size;
	
	//
	// get the size entry 
	// Set the _Size parameter to the actual valuecount available
	//
	_Size = *(int*)_pData;

	//
	// if we try to get more data than availible, 
	// we just fail
	//
	if( SetSize > _Size ) 
		return false;

	//
	// place pointer behind count
	//
	_pData = (destT*)((char*)_pData + sizeof( int ));


	// 
	// otherwise the memory is copied
	//
	CopyTypedMemory( _pData, _pValue, SetSize );

	// cannot fail any more
	return true;

} // SetTypedArray




//-------------------------------------------------------------------------------------------------
// Herkules
// utility template function
//
// 19981220	Herkules	_pValue is now a reference of a pointer. T and the data at _pData 
//						should be of same type in case _pValue==NULL. No further check can
//						be made here
//

template <class srcT, class destT> bool
GetTypedArray( srcT* _pData, destT*& _pValue, int& _Size )
{ 

	// if someone retrieved immediate access to our internal data (e.g. with a 
	// NULL  pointer in GetValue()), _pData and _pValue may be identical
	// in this case, there is nothing to move at all
	// This can only be true if the user GetValue()s the same variable twice,
	// first time with NULL, second time with the pointer obtained from first call!
	if( (void*)_pData == (void*)_pValue )
		return true;


	// remember size to get
	int GetSize = _Size;
	
	//
	// get and skip the size entry 
	// Set the _Size parameter to the actual valuecount available
	//
	_Size = *(int*)_pData;
	_pData = (srcT*)((char*)_pData + sizeof( int ));

	//
	// if we try to get more data than availible, 
	// we just fail
	//
	if( GetSize > _Size ) 
		return false;

	//
	// a NULL destination pointer means we jst have to return 
	// the pointer to the data instead of copying it
	//

	//
	// if we come here with NULL destination pointer, somebody else
	// checked that type of data and type of pointer are the same
	//
	if( _pValue == NULL )
	{
		_pValue = (destT*)_pData;
		return true;
	}

	// 
	// otherwise the memory is copied
	//
	CopyTypedMemory( _pValue, _pData, GetSize );

	return true;
} // GetTypedArray





//-------------------------------------------------------------------------------------------------
// utility template function
// take a string and make a typed value out of it
//

// we need a specialized operator to resolve ambiguities
istream& operator >> ( istream& _Str, LongInt& _Value )
{
	
	unsigned long highl;
	unsigned long lowl;

	// LongInt string has to be of format "0xffffffffffffffff"
	// Has to start with '0x', has to have 16 digits!

	string s;

	// read the string
	{
		_Str.width( 11 );
		_Str >> s;
		istringstream is( s );

		is >> hex >> highl;
	}
	
	{
		_Str.width( 9 );
		_Str >> s;
		istringstream is( s );

		is >> hex >> lowl;
	}

	// build the longint
	_Value = (LongInt)highl << 32 | lowl; 
	
	return _Str;
}



template <class T> bool
SingleValueFromString( T& _Data, const string& _String )
{
	// define a input stream and set to _String 
	istringstream is( _String );

	// out from the stream to the variable
	is >> _Data;

	return true;

} // SingleValueFromString





//-------------------------------------------------------------------------------------------------
// utility template function
//

// we need a specialized operator to resolve ambiguities
ostream& operator <<( ostream& _Str, const LongInt _Value )
{
	ostringstream os;

	unsigned long hl = _Value >> 32;
	unsigned long ll = _Value & 0xffffffff;

	// NOT CORRECT!

	os << hex << showbase << setw(8) << setfill( '0' ) << hl << noshowbase << setw(8) << ll;

	_Str << os.str();
	return _Str;
}




template <class T> bool
StringFromSingleValue( T _Data, string& _String )
{
	// define a output stream 
	ostringstream os;

	// out from the stream to the variable
	os << (T)_Data;

	_String = os.str();

	return true;

} // StringFromSingleValue







//-------------------------------------------------------------------------------------------------
//
//
// implementation
//
//
//-------------------------------------------------------------------------------------------------


//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************


#define GET_COMPILED_ONLY(DATAPTR,OBJPTR)					\
	unsigned char* DATAPTR = OBJPTR->FlushedObjectData();	\
	if( ! DATAPTR )		return false;				


template <class V> class TUncompiledValue : public IUncompiledValue 
{
public:
	

	//-------------------------------------------------------------------------
	template <class S> TUncompiledValue( CStoredObjectMemory* _pObj, CStoredVar* _pVar, const S* _Array, const int _Size ) 
	{
		m_pVar = _pVar;
		m_pObj = _pObj;

		m_Size = _Size;
		if( _Size > 1 ) 
		{
			m_Array = new V[ m_Size ];
			CopyTypedMemory( m_Array, _Array, m_Size ); 
		}
		else
		{
			m_Value = (V)*_Array;
			m_Array = &m_Value;
		}
	}


	//-------------------------------------------------------------------------
	virtual ~TUncompiledValue()
	{
		if( m_Size > 1 )
			delete[] m_Array;
	}


	//-------------------------------------------------------------------------
	virtual int CompiledSize()
	{
		int Size = 0;
		int ValueBaseSize;
		
		if( STORED_BASETYPE( GetVar()->GetType() ) == IStoredVar::Reference )
			ValueBaseSize = m_pObj->SizeOfReference(); 
		else
			ValueBaseSize = sizeof( V );


		if( ISARRAYVAR( m_pVar->GetType() ) )
		{
			Size = sizeof( OBJECTSIZE );
			Size += m_Size * ValueBaseSize;
		}
		else
			Size = ValueBaseSize;

		return Size;
	}


	//-------------------------------------------------------------------------
	virtual unsigned char* Compile( unsigned char* _pData )
	{
		if( ISARRAYVAR( GetVar()->GetType() ) )
		{
			*((int*)_pData) = m_Size;
			_pData += sizeof( int );

			//
			// References need special treatment
			//
			if( STORED_BASETYPE( GetVar()->GetType() ) == IStoredVar::Reference )
			{
				//
				// compile a reference
				//
				IStoredObject** pObjArr = (IStoredObject**)m_Array;

				for( int i = 0; i < m_Size; ++ i )
					_pData = m_pObj->CompileReference( _pData, pObjArr[i] ); 
			}
			else
			{
				//
				// compile 'normal' data
				//
				CopyMemory( (PVOID)_pData, m_Array, m_Size * sizeof(V) );
				_pData += m_Size * sizeof(V);
			}

		}
		else
		{
			*(V*)_pData = m_Value;
			_pData += sizeof( V ); 
		}


		return _pData;
	}



	//
	// data
	//
	V	m_Value;
	V*	m_Array;
	int m_Size;


protected:
	TUncompiledValue() 
	{}

		
	// link back to the object that owns this value
	CStoredObjectMemory* m_pObj;

};





//-----------------------------------------------------------------------------CUncompiledValue_String
// Herkules
// Uncompiled values of string type
//

class CUncompiledValue_String : public IUncompiledValue
{
public:

	//-------------------------------------------------------------------------
	template <class S> CUncompiledValue_String( CStoredVar* _pVar, const S* _Array, const int _Size ) 
	{
		m_pVar = _pVar;

		assert( _Size == 1 );
		StringFromSingleValue( *_Array, m_Value );
		m_ReservedSize = m_Value.size();
	}


	//-------------------------------------------------------------------------
	CUncompiledValue_String( CStoredVar* _pVar, const string& _Value, const int _MaxSize ) 
	{
		m_pVar = _pVar;

		m_Value = _Value;

		if( _MaxSize == 0 )
			m_ReservedSize = _Value.size();
		else
		{
			assert( _MaxSize >= _Value.size() );
			m_ReservedSize = _MaxSize;
		}
	}


	//-------------------------------------------------------------------------
	virtual int CompiledSize()
	{
		return sizeof( OBJECTSIZE ) + m_ReservedSize + sizeof('\0');
	}


	//-------------------------------------------------------------------------
	virtual unsigned char* Compile( unsigned char* _pData )
	{
		int ChunkSize = *((int*)_pData) = m_ReservedSize + 1;
		_pData += sizeof( int );

		// Copy the string to the buffer. Take one byte more to include the trailing '\0'
		CopyMemory( (PVOID)_pData, m_Value.c_str(), m_Value.size() + 1 );

		_pData += ChunkSize;

		return _pData;
	}

private:

	// The string itself
	string m_Value;

	// Memory to be reserverd. Always > m_Value.size()
	int m_ReservedSize;

}; // CUncompiledValue_String





//-----------------------------------------------------------------------------IStoredVarMemory----
// Herkules
// Interface class for all StoredVar-Classes
//

class IStoredVarMemory : public CStoredVar 
{
public:
	IStoredVarMemory( const string& _Varname, const STOREDVAR_TYPE _Type, const STOREDVAR_ID _VarID ) :
	CStoredVar( _Varname, _Type, _VarID )
	{}

	virtual ~IStoredVarMemory() {}

	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const unsigned char* _pValues, const int _Size ) = 0;
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const short* _pValues, const int _Size ) = 0;
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const int* _pValues, const int _Size ) = 0;
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const __int64* _pValues, const int _Size ) = 0;
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const float* _pValues, const int _Size ) = 0;
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const double* _pValues, const int _Size ) = 0;
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const string* _pValues, const int _Size ) = 0;
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, IStoredObject** _pValues, const int _Size ) = 0;

	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, unsigned char*& _pValue, int& _Size ) = 0;
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, short*& _pValue, int& _Size ) = 0;
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, int*& _pValue, int& _Size ) = 0;
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, __int64*& _pValue, int& _Size ) = 0;
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, float*& _pValue, int& _Size ) = 0;
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, double*& _pValue, int& _Size ) = 0;
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, string*& _pValue, int& _Size ) = 0;
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, IStoredObject**& _pValue, int& _Size ) = 0;

	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, unsigned char* _pValue, int& _Size ) = 0;
	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, short* _pValue, int& _Size ) = 0;
	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, int* _pValue, int& _Size ) = 0;
	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, __int64* _pValue, int& _Size ) = 0;
	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, float* _pValue, int& _Size ) = 0;
	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, double* _pValue, int& _Size ) = 0;
	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, string* _pValue, int& _Size ) = 0;
	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, IStoredObject** _pValue, int& _Size ) = 0;

};








//-----------------------------------------------------------------------------_GetCompiledValue---
// Herkules
// Helper template for TStoredVar
//

template <class srcT, class dstT>
static bool 
_GetCompiledValue( const srcT& _Dummy, CStoredVar* _pVar, CStoredObjectMemory* _pObj, dstT*& _pValue, int& _Size )
{
	GET_COMPILED_ONLY( pData, _pObj )

	srcT* pSrcValue = (srcT*)FindCompiledValue( _pVar, pData );

	if( pSrcValue )
	if( ISARRAYVAR( _pVar->GetType() ) )
	{
		return GetTypedArray( pSrcValue, _pValue, _Size );
	}
	else
	{
		return AssignSingleValueToArray( pSrcValue, _pValue, _Size );
	}

	return false;
}




//-----------------------------------------------------------------------------TStoredVar----------
// Herkules
// IStoredVarMemory is implemented with a template for some data types
//

template <class T> 
class TStoredVar : public IStoredVarMemory
{
public:

	//-------------------------------------------------------------------------
	TStoredVar( const string& _Varname, const STOREDVAR_TYPE _Type, const STOREDVAR_ID _VarID ) :
	IStoredVarMemory( _Varname, _Type, _VarID )
	{}


	//-------------------------------------------------------------------------
	static CStoredVar* Create( const string& _Varname, const STOREDVAR_TYPE _Type, const STOREDVAR_ID _VarID ) 
	{
		return new TStoredVar( _Varname, _Type, _VarID ); 
	}

	
	//
	// adding new values
	//

	//-------------------------------------------------------------------------
	#define CREATEFUNCTION(TYPE)																		\
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const TYPE* _pValues, const int _Size )	\
	{																									\
		IUncompiledValue* pVal = new TUncompiledValue<T>( _pObj, this, _pValues, _Size );				\
		_pObj->AddUncompiledValue( pVal );																\
		return true;																					\
	}									

	CREATEFUNCTION( unsigned char	) 
	CREATEFUNCTION( short			) 
	CREATEFUNCTION( int				) 
	CREATEFUNCTION( __int64			) 
	CREATEFUNCTION( float			) 
	CREATEFUNCTION( double			) 
	#undef CREATEFUNCTION

	//-------------------------------------------------------------------------
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, IStoredObject** _pValues, const int _Size )	
	{	
		return false;
	}

	//-------------------------------------------------------------------------
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const string* _pValues, const int _Size )	
	{	
		T Value;
		if( SingleValueFromString( Value, *_pValues ) )
		{
			IUncompiledValue* pVal = new TUncompiledValue<T>( _pObj, this, &Value, 1 );
			_pObj->AddUncompiledValue( pVal );
			return true;
		}
		return false;
	}


	//
	// getting values
	//

	
	// _GetCompiledValue needs a dummy variable in order to instantiate the template correctly
	T m_Dummy;

	//-------------------------------------------------------------------------CREATEFUNCTION------
	#define CREATEFUNCTION(TYPE)															\
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, TYPE*& _pValue, int& _Size )	\
	{																						\
		return _GetCompiledValue<T,TYPE>( m_Dummy, this, _pObj, _pValue, _Size );			\
	}


	CREATEFUNCTION( unsigned char	) 
	CREATEFUNCTION( short			) 
	CREATEFUNCTION( int				) 
	CREATEFUNCTION( __int64			) 
	CREATEFUNCTION( float			) 
	CREATEFUNCTION( double			) 
	#undef CREATEFUNCTION
	
	//-------------------------------------------------------------------------GetCompiledValue----
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, string*& _pValue, int& _Size )
	{
		GET_COMPILED_ONLY( pData, _pObj )
		
		if( ISARRAYVAR( GetType() ) )
			return false;

		T* pSrcValue = (T*)FindCompiledValue( this, pData );

		if( GetType() == IStoredVar::Byte ) 
		{
			//
			// special treatment for char variables, bc. standard
			// string conversion seems not to be OK for that
			//
			int ProxyValue = (int)*pSrcValue;
			return StringFromSingleValue( ProxyValue, *_pValue );
		}
		else
		{
			return StringFromSingleValue( *pSrcValue, *_pValue );
		}
	
	}


	//-------------------------------------------------------------------------GetCompiledValue----
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, IStoredObject**& _pValue, int& _Size )
	{
		GET_COMPILED_ONLY( pData, _pObj )

		// only the size is desired?
		if( ( _pValue != NULL ) & ( _Size == NULL ) )
		{
			int* pSrcValue = (int*)FindCompiledValue( this, pData );

			if( pSrcValue )
			{
				if( ISARRAYVAR( GetType() ) )
					_Size = *pSrcValue;
				else
					_Size = 1;
				return true;
			}
		}
		return false;
	}



	//
	// setting values
	//

	//-------------------------------------------------------------------------CREATEFUNCTION------
	#define CREATEFUNCTION(TYPE)																\
	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, TYPE* _pValue, int& _Size )		\
	{																							\
		GET_COMPILED_ONLY( pData, _pObj )														\
																								\
		T* pDstValue = (T*)FindCompiledValue( this, pData );									\
																								\
		if( pDstValue )																			\
		{																						\
			if( ISARRAYVAR( GetType() ) )														\
			{																					\
				return SetTypedArray( pDstValue, _pValue, _Size );								\
			}																					\
			else																				\
			{																					\
				return AssignSingleValueToArray( _pValue, pDstValue, _Size );					\
			}																					\
		}																						\
		return false;																			\
	}

	CREATEFUNCTION( unsigned char	) 
	CREATEFUNCTION( short			) 
	CREATEFUNCTION( int				) 
	CREATEFUNCTION( __int64			) 
	CREATEFUNCTION( float			) 
	CREATEFUNCTION( double			) 
	#undef CREATEFUNCTION


	//-------------------------------------------------------------------------SetCompiledValue----
	virtual bool 
	SetCompiledValue( CStoredObjectMemory* _pObj, string* _pValue, int& _Size )
	{
		GET_COMPILED_ONLY( pData, _pObj )
		
		if( _Size != 1 ) 
			return false;

		// cannot assign strings to any array type
		if( ISARRAYVAR( GetType() ) )
			return false;

		T* pDstValue = (T*)FindCompiledValue( this, pData );

		if( pDstValue )
		if( GetType() == IStoredVar::Byte ) 
		{
			//
			// special treatment for char variables, bc. standard
			// string conversion seems not to be OK for that
			//
			
			int ProxyValue;
			bool ret = SingleValueFromString( ProxyValue, *_pValue );
			*pDstValue = (unsigned char)ProxyValue;
			return ret;
		}
		else
		{
			return SingleValueFromString( *pDstValue, *_pValue );
		}
		return false;
	}


	//-------------------------------------------------------------------------SetCompiledValue----
	virtual bool 
	SetCompiledValue( CStoredObjectMemory* _pObj, IStoredObject** _pValue, int& _Size )
	{
		// cannot do anything for references, sorry
		return false;
	}


}; // TStoredVar




//-----------------------------------------------------------------------------_GetCompiledStringValue
// Herkules
// Helper template
// Convert a string from the data to the required value
// 

template <class T> 	
static bool 
_GetCompiledStringValue( CStoredVar* _pVar, CStoredObjectMemory* _pObj, T*& _pValue, int& _Size )
{

	GET_COMPILED_ONLY( pData, _pObj )

	if( _Size > 1 )
		return false;

	char* pSrcValue = (char*)FindCompiledValue( _pVar, pData );

	if( pSrcValue )
	{
		// skip the size entry 
		pSrcValue += sizeof( int );
		
		if( _Size == 1 )
		{

			SingleValueFromString( *_pValue, (string)pSrcValue );
			return true;
		}
		else
		{
			bool IsSize0 = ( _Size == 0 );

			// there only is one value!
			_Size = 1;

			return IsSize0;
		}
	}
	return false;
}



template <class T> 	
static bool 
_SetCompiledValue( CStoredVar* _pVar, CStoredObjectMemory* _pObj, T* _pValue, int& _Size )
{
	GET_COMPILED_ONLY( pData, _pObj )
	
	// we can only assign single values
	if( _Size != 1 )
		return false;

	char* pDstValue = (char*)FindCompiledValue( _pVar, pData );


	// transform data into a string
	string DataString;
	StringFromSingleValue( *_pValue, DataString );

	// get the size and skip the size entry 
	int Size = *(int*)pDstValue;

	// does the new string fit?
	// (must be '>', not '>=', bc. we need to add a trailing '\0\'
	if( Size > DataString.length() )
	{
		pDstValue += sizeof( int );
		CopyMemory( pDstValue, DataString.c_str(), DataString.length()+1 );

		return true;
	}
	else
	{
		return false;
	}

}



 

//-----------------------------------------------------------------------------CStoredVarMem_String
// Herkules
// IStoredVarMemory implementation for strings
//

class CStoredVarMem_String : public IStoredVarMemory
{
public:

	//-------------------------------------------------------------------------
	CStoredVarMem_String( const string& _Varname, const STOREDVAR_TYPE _Type, const STOREDVAR_ID _VarID ) :
	IStoredVarMemory( _Varname, IStoredVar::String, _VarID )
	{}


	//-------------------------------------------------------------------------
	static CStoredVar* Create( const string& _Varname, const STOREDVAR_TYPE _Type, const STOREDVAR_ID _VarID ) 
	{
		return new CStoredVarMem_String( _Varname, _Type, _VarID ); 
	}



	//-------------------------------------------------------------------------
	#define CREATEFUNCTION(TYPE)																		\
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const TYPE* _pValues, const int _Size )	\
	{																									\
		if( _Size > 1 ) return false;																	\
		IUncompiledValue* pVal = new CUncompiledValue_String( this, _pValues, 1 );						\
		_pObj->AddUncompiledValue( pVal );																\
		return true;																					\
	}										
	CREATEFUNCTION( unsigned char	) 
	CREATEFUNCTION( short			) 
	CREATEFUNCTION( int				) 
	CREATEFUNCTION( __int64			) 
	CREATEFUNCTION( float			) 
	CREATEFUNCTION( double			) 
	#undef CREATEFUNCTION

	//-------------------------------------------------------------------------
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const string* _pValues, const int _Size )	
	{	
		IUncompiledValue* pVal =  new CUncompiledValue_String( this, *_pValues, _Size ); 									
		_pObj->AddUncompiledValue( pVal );
		return true;
	}


	//-------------------------------------------------------------------------
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, IStoredObject** _pValues, const int _Size )	
	{	
		return false;
	}



	//-------------------------------------------------------------------------
	#define CREATEFUNCTION(TYPE)														\
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, TYPE*& _pValue, int& _Size )	\
	{																					\
		return _GetCompiledStringValue( this, _pObj, _pValue, _Size );					\
	}

	CREATEFUNCTION( unsigned char	) 
	CREATEFUNCTION( short			) 
	CREATEFUNCTION( int				) 
	CREATEFUNCTION( __int64			) 
	CREATEFUNCTION( float			) 
	CREATEFUNCTION( double			) 
//	CREATEFUNCTION( string			)	// can be optimized!!
	#undef CREATEFUNCTION


	// CStoredVarMem_String::GetCompiledValue for strings
	virtual bool 
	GetCompiledValue( CStoredObjectMemory* _pObj, string*& _pValue, int& _Size )
	{
		GET_COMPILED_ONLY( pData, _pObj )

		// the caller of this functio make sure that _pValue != NULL 
		// and _Size == 1
		assert( _pValue );
		assert( _Size == 1 );


		int* pSrcValue = (int*)FindCompiledValue( this, pData );

		if( pSrcValue )
		{
			// skip the count (max stringlen)
			char* pStringData = (char*)(++pSrcValue);

			// use a copy constructor of std::string to set the value
			*_pValue = pStringData;

			return true;
		}

		return false;
	}


	// CStoredVarMem_String::GetCompiledValue for referencearrays
	virtual bool 
	GetCompiledValue( CStoredObjectMemory* _pObj, IStoredObject**& _pValue, int& _Size )
	{
		GET_COMPILED_ONLY( pData, _pObj )

		// only the size is desired?
		if( ( _pValue != NULL ) & ( _Size == NULL ) )
		{
			int* pSrcValue = (int*)FindCompiledValue( this, pData );

			if( pSrcValue )
			{
				_Size = 1;
				return true;
			}
		}
		return false;
	}






	//
	// setting values
	//

	//-------------------------------------------------------------------------
	#define CREATEFUNCTION(TYPE)														\
	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, TYPE* _pValue, int& _Size )	\
	{																					\
		return _SetCompiledValue( this, _pObj, _pValue, _Size );						\
	}

	CREATEFUNCTION( unsigned char	) 
	CREATEFUNCTION( short			) 
	CREATEFUNCTION( int				) 
	CREATEFUNCTION( __int64			) 
	CREATEFUNCTION( float			) 
	CREATEFUNCTION( double			) 
//	CREATEFUNCTION( string			)	// can be optimized!!
	#undef CREATEFUNCTION

	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, string* _pValue, int& _Size )
	{
		GET_COMPILED_ONLY( pData, _pObj )

		// the caller of this functio make sure that _pValue != NULL 
		// and _Size == 1
		assert( _pValue );
		assert( _Size == 1 );

		int* pDstValue = (int*)FindCompiledValue( this, pData );

		if( pDstValue )
		{
			int MaxSize = *pDstValue;

			// skip the count (max stringlen)
			char* pStringData = (char*)(++pDstValue);

			// >= bc. there has to be the '\0' appended
			if( _pValue->length() >= MaxSize )
				return false;
			
			CopyMemory( pStringData, _pValue->c_str(), _pValue->length() + 1 );

		}
		return false;
	}


	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, IStoredObject** _pValue, int& _Size )
	{
		// cannot do anything for references, sorry
		return false;
	}



};




//-----------------------------------------------------------------------------CStoredVarMem_Ref---
// Herkules
// IStoredVarMemory implementation for references
//

class CStoredVarMem_Ref : public IStoredVarMemory
{
public:

	//-------------------------------------------------------------------------
	CStoredVarMem_Ref( const string& _Varname, const STOREDVAR_TYPE _Type, const STOREDVAR_ID _VarID ) :
	IStoredVarMemory( _Varname, IStoredVar::ReferenceArray, _VarID )
	{}


	//-------------------------------------------------------------------------
	static CStoredVar* Create( const string& _Varname, const STOREDVAR_TYPE _Type, const STOREDVAR_ID _VarID ) 
	{
		return new CStoredVarMem_Ref( _Varname, _Type, _VarID ); 
	}

	//-------------------------------------------------------------------------
	#define CREATEFUNCTION(TYPE)																		\
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, const TYPE* _pValues, const int _Size )	\
	{																									\
		return false;																					\
	}									

	CREATEFUNCTION( unsigned char	) 
	CREATEFUNCTION( short			) 
	CREATEFUNCTION( int				) 
	CREATEFUNCTION( __int64			) 
	CREATEFUNCTION( float			) 
	CREATEFUNCTION( double			) 
	CREATEFUNCTION( string			) 
	#undef CREATEFUNCTION



	//-------------------------------------------------------------------------
	virtual bool CreateNewValue( CStoredObjectMemory* _pObj, IStoredObject** _pValues, const int _Size )	
	{	
		IUncompiledValue* pVal = new TUncompiledValue<IStoredObject*>( _pObj, this, _pValues, _Size ); 
		_pObj->AddUncompiledValue( pVal );
		return true;
	}



	//-------------------------------------------------------------------------
	#define CREATEFUNCTION(TYPE)														\
	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, TYPE*& _pValue, int& _Size )	\
	{																					\
		GET_COMPILED_ONLY( pData, _pObj )												\
																						\
		if( ( _pValue != NULL ) & ( _Size == NULL ) )									\
		{																				\
			int* pSrcValue = (int*)FindCompiledValue( this, pData );					\
																						\
			if( pSrcValue )																\
			{																			\
				_Size = *pSrcValue;														\
				return true;															\
			}																			\
		}																				\
		return false;																	\
	}

	CREATEFUNCTION( unsigned char	) 
	CREATEFUNCTION( short			) 
	CREATEFUNCTION( int				) 
	CREATEFUNCTION( __int64			) 
	CREATEFUNCTION( float			) 
	CREATEFUNCTION( double			) 
	CREATEFUNCTION( string			) 
	#undef CREATEFUNCTION


	virtual bool GetCompiledValue( CStoredObjectMemory* _pObj, IStoredObject**& _pValue, int& _Size )
	{
		
		GET_COMPILED_ONLY( pData, _pObj )

		// this only works for existing arrays!
		if( ! _pValue ) 
			return false;
		
		// check wether the value is available
		unsigned char* pSrcValue = (unsigned char*)FindCompiledValue( this, pData );
		
		// no, it is not
		if( ! pSrcValue )	
			return false;
		
		

		// remember size to get
		int GetSize = _Size;
		
		//
		// get and skip the size entry 
		// Set the _Size parameter to the actual valuecount available
		//
		_Size = *(int*)pSrcValue;
		pSrcValue += sizeof( int );


		//
		// if we try to get more data than available, we just fail
		//
		if( GetSize > _Size ) 
			return false;


		// pData now points to an array of references.
		// the references class it self may have been defined by a derived driver,
		// so we only may assume a IReference interface

		// create the object vector
		for( int i = 0; i < GetSize; ++i )
		{

			_pValue[i] = _pObj->CreateFromReferenceData( (void*)pSrcValue ); 

			pSrcValue += _pObj->SizeOfReference();
		}

		return true;
	}



	//
	// seeting values
	//


	//-------------------------------------------------------------------------
	#define CREATEFUNCTION(TYPE)														\
	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, TYPE* _pValue, int& _Size )	\
	{																					\
		return false;																	\
	}

	CREATEFUNCTION( unsigned char	) 
	CREATEFUNCTION( short			) 
	CREATEFUNCTION( int				) 
	CREATEFUNCTION( __int64			) 
	CREATEFUNCTION( float			) 
	CREATEFUNCTION( double			) 
	CREATEFUNCTION( string			)
	#undef CREATEFUNCTION



	virtual bool SetCompiledValue( CStoredObjectMemory* _pObj, IStoredObject** _pValue, int& _Size )
	{
		GET_COMPILED_ONLY( pData, _pObj )

		// locate value in block
		unsigned char* pDstValue = (unsigned char*)FindCompiledValue( this, pData );

		// not found?
		if( ! pDstValue )	
			return false;

		int SetSize = _Size;

		// get the size entry 
		// Set the _Size parameter to the actual valuecount available
		_Size = *(int*)pDstValue;


		// if we try to get more data than available, we just fail
		if( SetSize > _Size ) 
			return false;

		// place pointer behind count
		pDstValue += sizeof( int );

		//
		// now set all reference data again to the array
		//
		for( int i = 0; i < SetSize; ++ i )
			pDstValue = _pObj->CompileReference( (unsigned char*)pDstValue, (IStoredObject*)_pValue[i] ); 

		return true;

	}

};



//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************
//*************************************************************************************************








//-------------------------------------------------------------------------------------------------
// Herkules
// look wether a certain variable exists in an object
//

bool 
CStoredObjectMemory::HasValue( IStoredVar* _pVar )
{
	// only compiled objects can be asked for values
	//
	COMPILEDONLY

	return ( FindCompiledValue( (CStoredVar*)_pVar, (unsigned char*)m_pCompiledObject ) != NULL );

} // HasValue












///////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//
// adding values
//
//
//
///////////////////////////////////////////////////////////////////////////////////////////////////


#define ADDVALUE(TYPE)																	\
bool CStoredObjectMemory::AddValue( IStoredVar* _pVar, TYPE _Value )					\
{																						\
	UNCOMPILEDONLY																		\
	return static_cast<IStoredVarMemory*>(_pVar)->CreateNewValue( this, &_Value, 1 );	\
} // AddValue


#define ADDVALUEARRAY(TYPE)																		\
bool CStoredObjectMemory::AddValue( IStoredVar* _pVar, const TYPE* _pValue, int _Size )			\
{																								\
	UNCOMPILEDONLY																				\
	return static_cast<IStoredVarMemory*>(_pVar)->CreateNewValue( this, _pValue, _Size );		\
}
				

ADDVALUE( unsigned char )
ADDVALUE( short )
ADDVALUE( int )
ADDVALUE( __int64 )
ADDVALUE( float )
ADDVALUE( double )


//
// array types
//
ADDVALUEARRAY( unsigned char )
ADDVALUEARRAY( short )
ADDVALUEARRAY( int )
ADDVALUEARRAY( __int64 )
ADDVALUEARRAY( float )
ADDVALUEARRAY( double )


//-------------------------------------------------------------------------------------------------
bool 
CStoredObjectMemory::AddValue( IStoredVar* _pVar, const string& _Value, int _MaxSize )
{
	UNCOMPILEDONLY
	return static_cast<IStoredVarMemory*>(_pVar)->CreateNewValue( this, &_Value, _MaxSize );
}


//-------------------------------------------------------------------------------------------------
// Herkules
// add an array of references to the storage
//

bool 
CStoredObjectMemory::AddValue( IStoredVar* _pVar, IStoredObject** _pObj, int _Size )
{
	UNCOMPILEDONLY
	return static_cast<IStoredVarMemory*>(_pVar)->CreateNewValue( this, _pObj, _Size );
}








///////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//
// setting values
//
//
//
///////////////////////////////////////////////////////////////////////////////////////////////////

#define SETVALUE(TYPE)																		\
bool CStoredObjectMemory::SetValue( IStoredVar* _pVar, TYPE _Value )						\
{																							\
	m_bIsFlushed = false;																	\
	int Size = 1;																			\
	return static_cast<IStoredVarMemory*>(_pVar)->SetCompiledValue( this, &_Value, Size );	\
}


#define SETVALUEARRAY(TYPE)																		\
bool CStoredObjectMemory::SetValue( IStoredVar* _pVar, TYPE* _pValue, int& _Size )				\
{																								\
	m_bIsFlushed = false;																		\
	return static_cast<IStoredVarMemory*>(_pVar)->SetCompiledValue( this, _pValue, _Size );		\
}


SETVALUE( unsigned char )
SETVALUE( short )
SETVALUE( int )
SETVALUE( __int64 )
SETVALUE( float )
SETVALUE( double )

//
// arrays
//
SETVALUEARRAY( unsigned char )
SETVALUEARRAY( short )
SETVALUEARRAY( int )
SETVALUEARRAY( __int64 )
SETVALUEARRAY( float )
SETVALUEARRAY( double )

//
// string types
//

//-------------------------------------------------------------------------------------------------
bool 
CStoredObjectMemory::SetValue( IStoredVar* _pVar, const string& _Value )
{
	m_bIsFlushed = false;

	int Size = 1;

	return static_cast<IStoredVarMemory*>(_pVar)->SetCompiledValue( this, &const_cast<string&>(_Value), Size );
}


//-------------------------------------------------------------------------------------------------
// Herkules
// set an array of references
//

bool 
CStoredObjectMemory::SetValue( IStoredVar* _pVar, IStoredObject** _pObj, int& _Size )
{
	m_bIsFlushed = false;
	return static_cast<IStoredVarMemory*>(_pVar)->SetCompiledValue( this, _pObj, _Size );
}







///////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//
// getting values
//
//
//
///////////////////////////////////////////////////////////////////////////////////////////////////


#define GETVALUE(TYPE)																		\
bool CStoredObjectMemory::GetValue( IStoredVar* _pVar, TYPE& _Value )						\
{																							\
	int Size = 1;																			\
	TYPE* Pointer = &_Value;																\
	return static_cast<IStoredVarMemory*>(_pVar)->GetCompiledValue( this, Pointer, Size );	\
} // GetValue



// If the pointer is NULL, the user wants to access data directly by getting
// our internal pointer to the data
// So we have to check wether types are compatible upfront 

#define GETVALUEARRAY(TYPE, VARTYPECONST)														\
bool CStoredObjectMemory::GetValue( IStoredVar* _pVar, TYPE*& _pValue, int& _Size )				\
{																								\
	if( (!_pValue) && ( _pVar->GetType() &~VARTYPEARRAYMASK ) != VARTYPECONST )					\
		return false;																			\
	return static_cast<IStoredVarMemory*>(_pVar)->GetCompiledValue( this, _pValue, _Size );		\
} // GetValue


GETVALUE( unsigned char )
GETVALUE( short )
GETVALUE( int )
GETVALUE( __int64 )
GETVALUE( float )
GETVALUE( double )

//
// arrays
//
GETVALUEARRAY( unsigned char, IStoredVar::Byte )
GETVALUEARRAY( short, IStoredVar::Short )
GETVALUEARRAY( int, IStoredVar::Int )
GETVALUEARRAY( __int64, IStoredVar::LongInt )
GETVALUEARRAY( float, IStoredVar::Float )
GETVALUEARRAY( double, IStoredVar::Double )



//-------------------------------------------------------------------------------------------------
//
// access a value into a string
//
bool 
CStoredObjectMemory::GetValue( IStoredVar* _pVar, string& _Value )
{
	int Size = 1;
	string* Pointer = &_Value;

	return static_cast<IStoredVarMemory*>(_pVar)->GetCompiledValue( this, Pointer, Size );
}



//--------------------------------------------------------------------------------------------------
// Herkules
// Get a reference type value
//

bool 
CStoredObjectMemory::GetValue( IStoredVar* _pVar, IStoredObject** _pObj, int& _Size )
{

	// the call can only work with reference variables
	if( ( _pVar->GetType() & ~VARTYPEARRAYMASK ) != IStoredVar::Reference )
		return false;

	return static_cast<IStoredVarMemory*>(_pVar)->GetCompiledValue( this, _pObj, _Size );
}


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//
// end of access routines
//
//
///////////////////////////////////////////////////////////////////////////////////////////////////



//-----------------------------------------------------------------------------ReAllocate----------
// Herkules
// Throw away the whole content of the object and take it to the state just after AddObject() 
//

bool 
CStoredObjectMemory::ReAllocate()
{
	
	// an object can only be ReAllocated if it has been flushed once
	if( ! m_bIsFlushed )
		return false;
	
	// free the memory, if allocated
	DeAllocateCompiledObject();

	// set the objects data size to 0
	m_ObjectSize = 0;

	// no, it's not flushed
	m_bIsFlushed = false;

	// nothing can go wrong here
	return true;

} // ReAllocate




//-----------------------------------------------------------------------------SizeOfReference-----
// Herkules
// in order to allow derived classes to use the Compile() coded here, we 
// virtualize the size of a reference structure 
//
int
CStoredObjectMemory::SizeOfReference()
{
	return sizeof( REFERENCEDATAMEMORY );
}



//-----------------------------------------------------------------------------FlushedObjectSize---
// Herkules
// Calculate size of compiled data. If it is not yet compiled, calculate size
// from the blown-up structure
//

OBJECTSIZE 
CStoredObjectMemory::FlushedObjectSize()
{


	if( m_pCompiledObject )
	{
		return m_ObjectSize;
	}
	else
	{

		std::list<IUncompiledValue*>::iterator It;

		
		// size of storage needed
		int Size = 0;

		
		// number of values
		int ValueCount = m_lUncompiledValues.size();


		// this first value is the number of values in the object
		Size += sizeof( STOREDVAR_ID );


		// now make place for all _CompiledHeaderDirectoryEntry
		Size += ValueCount * sizeof( _CompiledHeaderDirectoryEntry );
		

		//
		// now the space for all values
		//
		for( It = m_lUncompiledValues.begin(); It != m_lUncompiledValues.end(); ++It )
		{
			Size += (*It)->CompiledSize();
		}
		

		// Size is calculated here
		return Size;
		
	}

} // FlushedObjectSize





//-----------------------------------------------------------------------------CompileReference----
// Herkules
// virtual function compiling a reference into the compiled memory block.
// overloading this function allows derived classes to use our Compile() method
//

unsigned char* 
CStoredObjectMemory::CompileReference( unsigned char* _pData, IStoredObject* _pObj )
{

	REFERENCEDATAMEMORY Ref;

	Ref.m_pObject = _pObj;

	*((REFERENCEDATAMEMORY*)_pData) = Ref;
	_pData += SizeOfReference();

	return _pData;

} // CompileReference




//-----------------------------------------------------------------------------Compile-------------
// Herkules
// Take the uncompiled list all values and pack them into a single memory block of 
// fixed size
//

bool 
CStoredObjectMemory::Compile( )
{

	// locals 
	std::list<IUncompiledValue*>::iterator It;


	//
	// if the object is already compiled, we cannot do anything
	//
	if( m_pCompiledObject )		return false;



	//
	// calculate size and remember it
	//
	m_ObjectSize = FlushedObjectSize();


	//
	// now m_ObjectSize contains the actual size of the buffer needed (in byte)
	//
	m_pCompiledObject = AllocateCompiledObject( m_ObjectSize );


	
	//
	// ok, now compile data to that buffer
	//
	
	
	//
	// define two pointers. One pointing to the next directory entry and one to 
	// the next data position
	//
	_CompiledHeaderDirectoryEntry* pDir = (_CompiledHeaderDirectoryEntry*)(m_pCompiledObject + sizeof( STOREDVAR_ID ));

	

	// number of values in object
	int ValueCount = m_lUncompiledValues.size();


	// directory starts with number of values
	*(STOREDVAR_ID*)m_pCompiledObject = ValueCount;
	

	// data starts after directory
	unsigned char* pData = (unsigned char*)pDir + ValueCount * sizeof( _CompiledHeaderDirectoryEntry );

	//
	// for all values, write them to the buffer
	//
	for( It = m_lUncompiledValues.begin(); It != m_lUncompiledValues.end(); ++It )
	{
		// locals

		pDir->ID = (*It)->GetVar()->GetID();
		pDir->Offset = pData - (unsigned char*)pDir ;

		pDir++;

		pData = (*It)->Compile( pData );

		delete (*It);
	}

	m_lUncompiledValues.clear();


	// flag the we are flushed now!
	m_bIsFlushed = true;


	// done 
	return true;

} // Compile()






//---------------------------------------------------------------------------------------------
// use data from an external source
// this is a service for derived classes
//

void 
CStoredObjectMemory::AssignData( unsigned char* pData, OBJECTSIZE _Size )
{
	assert( ! m_pCompiledObject );

	// pData MUST be allocated with AllocateComiledObject
	m_pCompiledObject = pData;
	m_ObjectSize = _Size;

}




//---------------------------------------------------------------------------------------------
// Herkules
// create an object from raw reference data
//

IStoredObject* 
CStoredObjectMemory::CreateFromReferenceData( void* _pRefDat )
{

	return ((REFERENCEDATAMEMORY*)_pRefDat)->m_pObject;


} // CreateFromReferenceData










//-------------------------------------------------------------------------------------------------
//
// CStoredClassMemory                                                          CStoredClassMemory
//
//-------------------------------------------------------------------------------------------------


//-----------------------------------------------------------------------------CStoredClassMemory--
// Herules
// Constructor
//

CStoredClassMemory::CStoredClassMemory( const string& _Classname, 
										const STOREDCLASS_ID _ClassID, 
										CStorage* _pStorage, 
										IStoredClass* const _pBaseClass ) :
CStoredClass( _Classname, _ClassID, _pStorage, _pBaseClass ),
m_NumObjects( 0 )
{
} 




//-----------------------------------------------------------------------------~CStoredClassMemory-
// Herkules
// Destructor
//

CStoredClassMemory::~CStoredClassMemory() 
{
	//
	// delete all objects we created 
	//
	int i;
	for( i = 0; i< m_NumObjects; i++ )
	{
		if( m_aObjects[i] )
			delete m_aObjects[i];
	}

	//
	// free the array itself
	//
	m_aObjects.clear();

}




//-----------------------------------------------------------------------------AssignObject--------
// Herkules
// Protected: take an object and add it to the classes list of objects
//

bool 
CStoredClassMemory::AssignObject( IStoredObject* _pObject )
{
/*
	//
	// in case of increase, double the size of the array
	//
	if( m_NumObjects+1 > m_aObjects.capacity() )
	{
		m_aObjects.resize( (m_NumObjects+1) * 2, NULL );	
		// remark: I first used m_aObjects.size() as a basis for the new size,
		// but this can be zero and m_NumObjects is the shorter path anyway
	}


	// store the pointer to the new object
	m_aObjects[m_NumObjects] = pO;
*/

	// Add the pointer to the array
	m_aObjects.push_back( static_cast<CStoredObjectMemory*>(_pObject) );

	// note that we have one more now!
	m_NumObjects++;

	// Nearly cannot fail!
	return true;

} // AssignObject


//-----------------------------------------------------------------------------AddObject-----------
// Herkules
// add a new object 
//

IStoredObject* 
CStoredClassMemory::AddObject( )
{

	//
	// create a new memory object.
	// the current counter is the index of that object
	//
	CStoredObjectMemory* pO = new CStoredObjectMemory( m_NumObjects, this);

	// Take it under the hood
	AssignObject( pO );

	// The base class gets the object as well! 
//	if( GetBaseClass() ) 
//		static_cast<CStoredClassMemory*>(GetBaseClass())->AssignObject( pO ); 

		
/*
	//
	// in case of increase, double the size of the array
	//
	if( m_NumObjects+1 > m_aObjects.capacity() )
	{
		m_aObjects.resize( (m_NumObjects+1) * 2, NULL );	
		// remark: I first used m_aObjects.size() as a basis for the new size,
		// but this can be zero and m_NumObjects is the shorter path anyway
	}


	// store the pointer to the new object
	m_aObjects[m_NumObjects] = pO;

	// note that we have one more now!
	m_NumObjects++;
*/

	// return the new object pointer
	return pO;

} // AddObject




//-----------------------------------------------------------------------------CreateVar-----------
// Herkules
// Create a SUITABLE CStoredClass-derived object
//

CStoredVar* 
CStoredClassMemory::CreateVar( const string& _Varname, STOREDVAR_TYPE _Type, STOREDVAR_ID _ID )
{

	CStoredVar* pVar;

	// this call provides 'old' functionality
//	pVar = super::CreateVar( _Varname, _Type, _ID );

	
	// but now we like to do it another way....

	typedef CStoredVar* (*VarCreateFuncPtr)(const string& _Varname, const STOREDVAR_TYPE _Type, const STOREDVAR_ID _ID) ;

	// create a table of create-functions
	static VarCreateFuncPtr CreatePtrArray[] =
	{
		NULL,
		TStoredVar<Byte>::Create,
		TStoredVar<Short>::Create,
		TStoredVar<Int>::Create,
		TStoredVar<LongInt>::Create,
		TStoredVar<Float>::Create,
		TStoredVar<Double>::Create,
		CStoredVarMem_String::Create,
		CStoredVarMem_Ref::Create
	} ;

	// call the right create function
	pVar = CreatePtrArray[ STORED_BASETYPE( _Type ) ]( _Varname, _Type, _ID );


	return pVar;
}




