// main.cpp
//**********************************************************************************************
//
//
//				(C) Copyright 1998 by Dipl. Phys. Joerg Plewe, HARDCODE Development
//					All rights reserved. Copying, modification,
//					distribution or publication without the prior written 
//					consent of the author is prohibited.
//
//
//	Filename:		main.cpp
//	Author:			Herkules
//  Created:		19990207
//  Version:		$Version$
//
//	Module:			StorageExamples
//  Dependecies:	$Dependencies$
//
//  History:
//  $Date$		$Who$		$What$
//
/*	Purpose:

Setup a simple database and fill in some values.
It will show how inheritance is used an can be exploited.


Inheritance in the storage is a pure REPOSITORY SERVICE! 
This means when iterating the objects of a base class, the objects of a derived class
will never appear. 

But for the repository can be queried for all classes, it is possible to all objects
of a base class (including derived classes objects) as a union of some iterations.

*/
//**********************************************************************************************


#include <iostream>
#include <assert.h>

#include "..\..\Storage\IStorage.h"




//
// define some names for classes
//
#define NAME_CLASS_BASE		"BaseClass"
#define NAME_CLASS_DERIVED	"DerivedClass"
#define NAME_CLASS_DUMMY	"DummyClass"


//
// Define some names for variables
// These variables are very nice for a base class, bc. they
// allow to identify an object in three ways
//
#define NAME_VAR_ID		"ID"
#define TYPE_VAR_ID		IStoredVar::Int

#define NAME_VAR_NAME	"Name"
#define TYPE_VAR_NAME	IStoredVar::String

#define NAME_VAR_POP	"POP"
#define TYPE_VAR_POP	IStoredVar::LongInt

// A special variable only for the derived class
#define NAME_VAR_X		"X"
#define TYPE_VAR_X		IStoredVar::Double



//-----------------------------------------------------------------------------IsDerived-----------
// Little helper function to find out wether a certain class is 
bool 
IsDerived( IStoredClass* _pBaseClass, IStoredClass* _pPossiblyDerived )
{

	IStoredClass* pTestClass = _pPossiblyDerived;

	// Get the ID of the target baseclass. It can be used for comparison of classes
	int BaseClassID = _pBaseClass->GetID();

	// Loop until either the pTestClass has the desired type or there are no more base classes
	while(		pTestClass->GetID() != BaseClassID  
			&&  (pTestClass = pTestClass->GetBaseClass()) )
	{}


	// Has a suitable class been found?
	return pTestClass != NULL; 

}







//-----------------------------------------------------------------------------main----------------
void main()
{

	//
	// locals 
	//	
	bool ret;

	// we will need an engine
	IStorageEngine* pEngine;

	// we do it with just one base- and one derived class
	IStoredClass*	pBaseClass;
	IStoredClass*	pDerivedClass;
	IStoredClass*	pDummyClass;


	// a couple of variables
	IStoredVar*		pVarID;
	IStoredVar*		pVarName;
	IStoredVar*		pVarPOP;
	IStoredVar*		pVarX;


	// Create the libary
	pEngine = IStorageEngine::New( IStorageEngine::Memory );
	
	// Connect: does nothing for a memory database
	ret = pEngine->Connect( "Inheritance.hsl" );


	// Add the base class to repository
	pBaseClass = pEngine->AddClass( NAME_CLASS_BASE );
	assert( pBaseClass );


	//
	// add variables to the class
	//
	pVarID		= pBaseClass->AddVar( NAME_VAR_ID,		TYPE_VAR_ID );
	pVarName	= pBaseClass->AddVar( NAME_VAR_NAME,	TYPE_VAR_NAME );
	pVarPOP		= pBaseClass->AddVar( NAME_VAR_POP,		TYPE_VAR_POP );

	
	//
	// Create some base objects
	//
	for( int i = 0; i < 10; ++ i )
	{
		IStoredObject* pBaseObj = pBaseClass->AddObject();
		assert( pBaseObj );

		// Fill the integer variable. The data comes from a string
		ret = pBaseObj->AddValue( pVarID, i );  

		// Fill the name variable. The data comes from a float :-)
		ret = pBaseObj->AddValue( pVarName, i  );  

		// Fill the string variable. The data comes from a float :-)
		ret = pBaseObj->AddValue( pVarName, CreatePOP()  );  

		// flush the object, make it ready for use
		ret = pBaseObj->Flush();

		// get rid of it for now
		ret = pBaseObj->Release();
	
	}




	//
	// Now add the derived class to repository
	// Note the second argument: this is where inheritance takes place!
	//
	pDerivedClass = pEngine->AddClass( NAME_CLASS_DERIVED, pBaseClass );
	assert( pDerivedClass );

	// We ask the derived class for the variables of the base class
	pVarID		= pDerivedClass->GetVar( NAME_VAR_ID );
	pVarName	= pDerivedClass->GetVar( NAME_VAR_NAME );
	pVarPOP		= pDerivedClass->GetVar( NAME_VAR_POP );

	// We add an additional variable for the derived class
	pVarX		= pDerivedClass->AddVar( NAME_VAR_X, TYPE_VAR_X );


	//
	// Create some derived objects
	//
	for( i = 0; i < 10; ++ i )
	{
		IStoredObject* pDerivedObj = pDerivedClass->AddObject();
		assert( pDerivedObj );

		// Fill the name variable. The data comes from a float :-)
		ret = pDerivedObj->AddValue( pVarName, i * 1.23456789f );  

		// Fill the variable X.
		ret = pDerivedObj->AddValue( pVarX, i * 1.23456789 );  


		// Flush the object, make it ready for use
		ret = pDerivedObj->Flush();

		// Get rid of it for now
		ret = pDerivedObj->Release();

	}




	//
	// Now we are going to query all objects of the base type, which means all
	// objects of pBaseClass + pDerivedClass
	//

	// We add another class to makes things more complicated and to see that is really works
	pDummyClass = pEngine->AddClass( NAME_CLASS_DUMMY );


	IStoredClassList* pClassList = pEngine->GetClassList();
	IStoredClass* pClass = pClassList->First();
	string Name;
	double X;



	//
	// The following loop looks a bit complicated.
	//
	// We are iterating on all possible classes. For each class we determine (with a little 
	// helper function) wether it is derived from the desired base class pBaseClass.
	// If it is, we iterate on that class and get all objects.
	//
	while( pClass )
	{

		//
		// Is pClass derived from pBaseClass?
		//
		if( IsDerived( pBaseClass, pClass ) )
		{

			//
			// Iterate all objects
			//
			IStoredObjectList* pList = pClass->GetObjectList();
			IStoredObject* pObj = pList->First();

			while( pObj )
			{

				// All objects should have a name from the base class
				ret = pObj->GetValue( pVarName, Name );

				// Only derived objects should have a X
				ret = pObj->GetValue( pVarX, X );
				

				// Forget the object and go to the next
				pObj->Release();
				pObj = pList->Next();
			}

			pList->Release();
		}

		// Next class from repository
		pClass = pClassList->Next();
	}
	



	
	// Don't forget that!
	pClassList->Release();

	// 
	// shutdown the library
	//
	pEngine->Close();
	pEngine->Release();


}